import pytest
from pytest_mock import MockerFixture
from synthex import Synthex
from transformers.trainer_utils import TrainOutput
from datasets import DatasetDict, Dataset
from unittest.mock import MagicMock
from typing import Any

from artifex.models.reranker import Reranker
from artifex.core import ParsedModelInstructions
from artifex.config import config


@pytest.fixture
def mock_dependencies(mocker: MockerFixture) -> None:
    """
    Fixture to mock external dependencies for Reranker.
    
    Args:
        mocker (MockerFixture): The pytest-mock fixture for mocking.
    """
    
    mocker.patch(
        'artifex.models.reranker.reranker.AutoModelForSequenceClassification.from_pretrained',
        return_value=MagicMock()
    )
    mocker.patch(
        'artifex.models.reranker.reranker.AutoTokenizer.from_pretrained',
        return_value=MagicMock()
    )
    mocker.patch('artifex.models.reranker.reranker.torch.cuda.is_available', return_value=False)
    mocker.patch('artifex.models.reranker.reranker.torch.backends.mps.is_available', return_value=False)


@pytest.fixture
def mock_synthex(mocker: MockerFixture) -> Synthex:
    """
    Fixture to create a mock Synthex instance.
    
    Args:
        mocker (MockerFixture): The pytest-mock fixture for mocking.
    
    Returns:
        Synthex: A mocked Synthex instance.
    """
    
    return mocker.MagicMock()


@pytest.fixture
def mock_get_model_output_path(mocker: MockerFixture) -> MagicMock:
    """
    Fixture to mock get_model_output_path utility function.
    
    Args:
        mocker (MockerFixture): The pytest-mock fixture for mocking.
    
    Returns:
        MagicMock: Mocked get_model_output_path function.
    """
    
    return mocker.patch(
        'artifex.models.reranker.reranker.get_model_output_path',
        return_value="/test/output/model"
    )


@pytest.fixture
def mock_training_args(mocker: MockerFixture) -> MagicMock:
    """
    Fixture to mock TrainingArguments.
    
    Args:
        mocker (MockerFixture): The pytest-mock fixture for mocking.
    
    Returns:
        MagicMock: Mocked TrainingArguments class.
    """
    
    return mocker.patch('artifex.models.reranker.reranker.TrainingArguments')


@pytest.fixture
def mock_silent_trainer(mocker: MockerFixture) -> MagicMock:
    """
    Fixture to mock SilentTrainer.
    
    Args:
        mocker (MockerFixture): The pytest-mock fixture for mocking.
    
    Returns:
        MagicMock: Mocked SilentTrainer class.
    """
    
    mock_trainer_instance = MagicMock()
    mock_trainer_instance.train.return_value = TrainOutput(
        global_step=100, training_loss=0.5, metrics={}
    )
    mock_trainer_class = mocker.patch(
        'artifex.models.reranker.reranker.SilentTrainer',
        return_value=mock_trainer_instance
    )
    return mock_trainer_class


@pytest.fixture
def mock_os_path_exists(mocker: MockerFixture) -> MagicMock:
    """
    Fixture to mock os.path.exists.
    
    Args:
        mocker (MockerFixture): The pytest-mock fixture for mocking.
    
    Returns:
        MagicMock: Mocked os.path.exists function.
    """
    
    return mocker.patch('artifex.models.reranker.reranker.os.path.exists', return_value=True)


@pytest.fixture
def mock_os_remove(mocker: MockerFixture) -> MagicMock:
    """
    Fixture to mock os.remove.
    
    Args:
        mocker (MockerFixture): The pytest-mock fixture for mocking.
    
    Returns:
        MagicMock: Mocked os.remove function.
    """
    
    return mocker.patch('artifex.models.reranker.reranker.os.remove')


@pytest.fixture
def reranker(
    mock_dependencies: None,
    mock_synthex: Synthex,
    mocker: MockerFixture
) -> Reranker:
    """
    Fixture to create a Reranker instance for testing.
    
    Args:
        mock_dependencies (None): Fixture that mocks external dependencies.
        mock_synthex (Synthex): A mocked Synthex instance.
        mocker (MockerFixture): The pytest-mock fixture for mocking.
    
    Returns:
        Reranker: A Reranker instance.
    """
    
    reranker = Reranker(synthex=mock_synthex)
    
    # Mock _build_tokenized_train_ds to return a valid DatasetDict
    mock_dataset = DatasetDict({
        "train": Dataset.from_dict({"query": ["q"], "document": ["d"], "labels": [1.0]}),
        "test": Dataset.from_dict({"query": ["q2"], "document": ["d2"], "labels": [2.0]})
    })
    mocker.patch.object(reranker, '_build_tokenized_train_ds', return_value=mock_dataset)
    
    return reranker


@pytest.mark.unit
def test_perform_train_pipeline_calls_build_tokenized_train_ds(
    reranker: Reranker,
    mock_get_model_output_path: MagicMock,
    mock_training_args: MagicMock,
    mock_silent_trainer: MagicMock,
    mock_os_path_exists: MagicMock,
    mock_os_remove: MagicMock,
    mocker: MockerFixture
) -> None:
    """
    Test that _perform_train_pipeline calls _build_tokenized_train_ds with correct arguments.
    
    Args:
        reranker (Reranker): The Reranker instance.
        mock_get_model_output_path (MagicMock): Mocked get_model_output_path function.
        mock_training_args (MagicMock): Mocked TrainingArguments class.
        mock_silent_trainer (MagicMock): Mocked SilentTrainer class.
        mock_os_path_exists (MagicMock): Mocked os.path.exists function.
        mock_os_remove (MagicMock): Mocked os.remove function.
        mocker (MockerFixture): The pytest-mock fixture for mocking.
    """
    
    mock_build = mocker.spy(reranker, '_build_tokenized_train_ds')
    
    user_instructions = ParsedModelInstructions(
        user_instructions=["healthcare"],
        language="english"
    )
    
    reranker._perform_train_pipeline(
        user_instructions=user_instructions,
        output_path="/output",
        num_samples=200,
        num_epochs=5
    )
    
    mock_build.assert_called_once_with(
        user_instructions=user_instructions,
        output_path="/output",
        num_samples=200,
        train_datapoint_examples=None
    )


@pytest.mark.unit
def test_perform_train_pipeline_passes_examples_to_build_tokenized(
    reranker: Reranker,
    mock_get_model_output_path: MagicMock,
    mock_training_args: MagicMock,
    mock_silent_trainer: MagicMock,
    mock_os_path_exists: MagicMock,
    mock_os_remove: MagicMock,
    mocker: MockerFixture
) -> None:
    """
    Test that _perform_train_pipeline passes examples to _build_tokenized_train_ds.
    
    Args:
        reranker (Reranker): The Reranker instance.
        mock_get_model_output_path (MagicMock): Mocked get_model_output_path function.
        mock_training_args (MagicMock): Mocked TrainingArguments class.
        mock_silent_trainer (MagicMock): Mocked SilentTrainer class.
        mock_os_path_exists (MagicMock): Mocked os.path.exists function.
        mock_os_remove (MagicMock): Mocked os.remove function.
        mocker (MockerFixture): The pytest-mock fixture for mocking.
    """
    
    mock_build = mocker.spy(reranker, '_build_tokenized_train_ds')
    
    user_instructions = ParsedModelInstructions(
        user_instructions=["healthcare"],
        language="english"
    )
    
    examples: list[dict[str, Any]] = [
        {"query": "test query", "document": "test doc", "score": 5.0}
    ]
    
    reranker._perform_train_pipeline(
        user_instructions=user_instructions,
        output_path="/output",
        train_datapoint_examples=examples
    )
    
    call_kwargs = mock_build.call_args[1]
    assert call_kwargs['train_datapoint_examples'] == examples


@pytest.mark.unit
def test_perform_train_pipeline_checks_cuda_availability(
    reranker: Reranker,
    mock_get_model_output_path: MagicMock,
    mock_training_args: MagicMock,
    mock_silent_trainer: MagicMock,
    mock_os_path_exists: MagicMock,
    mock_os_remove: MagicMock,
    mocker: MockerFixture
) -> None:
    """
    Test that _perform_train_pipeline checks CUDA availability.
    
    Args:
        reranker (Reranker): The Reranker instance.
        mock_get_model_output_path (MagicMock): Mocked get_model_output_path function.
        mock_training_args (MagicMock): Mocked TrainingArguments class.
        mock_silent_trainer (MagicMock): Mocked SilentTrainer class.
        mock_os_path_exists (MagicMock): Mocked os.path.exists function.
        mock_os_remove (MagicMock): Mocked os.remove function.
        mocker (MockerFixture): The pytest-mock fixture for mocking.
    """
    
    mock_cuda = mocker.patch('artifex.models.reranker.reranker.torch.cuda.is_available')
    
    user_instructions = ParsedModelInstructions(
        user_instructions=["healthcare"],
        language="english"
    )
    
    reranker._perform_train_pipeline(
        user_instructions=user_instructions,
        output_path="/output"
    )
    
    mock_cuda.assert_called_once()


@pytest.mark.unit
def test_perform_train_pipeline_checks_mps_availability(
    reranker: Reranker,
    mock_get_model_output_path: MagicMock,
    mock_training_args: MagicMock,
    mock_silent_trainer: MagicMock,
    mock_os_path_exists: MagicMock,
    mock_os_remove: MagicMock,
    mocker: MockerFixture
) -> None:
    """
    Test that _perform_train_pipeline checks MPS availability.
    
    Args:
        reranker (Reranker): The Reranker instance.
        mock_get_model_output_path (MagicMock): Mocked get_model_output_path function.
        mock_training_args (MagicMock): Mocked TrainingArguments class.
        mock_silent_trainer (MagicMock): Mocked SilentTrainer class.
        mock_os_path_exists (MagicMock): Mocked os.path.exists function.
        mock_os_remove (MagicMock): Mocked os.remove function.
        mocker (MockerFixture): The pytest-mock fixture for mocking.
    """
    
    mock_mps = mocker.patch('artifex.models.reranker.reranker.torch.backends.mps.is_available')
    
    user_instructions = ParsedModelInstructions(
        user_instructions=["healthcare"],
        language="english"
    )
    
    reranker._perform_train_pipeline(
        user_instructions=user_instructions,
        output_path="/output"
    )
    
    mock_mps.assert_called_once()


@pytest.mark.unit
def test_perform_train_pipeline_uses_pin_memory_when_cuda_available(
    reranker: Reranker,
    mock_get_model_output_path: MagicMock,
    mock_training_args: MagicMock,
    mock_silent_trainer: MagicMock,
    mock_os_path_exists: MagicMock,
    mock_os_remove: MagicMock,
    mocker: MockerFixture
) -> None:
    """
    Test that _perform_train_pipeline sets pin_memory when CUDA is available.
    
    Args:
        reranker (Reranker): The Reranker instance.
        mock_get_model_output_path (MagicMock): Mocked get_model_output_path function.
        mock_training_args (MagicMock): Mocked TrainingArguments class.
        mock_silent_trainer (MagicMock): Mocked SilentTrainer class.
        mock_os_path_exists (MagicMock): Mocked os.path.exists function.
        mock_os_remove (MagicMock): Mocked os.remove function.
        mocker (MockerFixture): The pytest-mock fixture for mocking.
    """
    
    mocker.patch('artifex.models.reranker.reranker.torch.cuda.is_available', return_value=True)
    
    user_instructions = ParsedModelInstructions(
        user_instructions=["healthcare"],
        language="english"
    )
    
    reranker._perform_train_pipeline(
        user_instructions=user_instructions,
        output_path="/output"
    )
    
    call_kwargs = mock_training_args.call_args[1]
    assert call_kwargs['dataloader_pin_memory'] is True


@pytest.mark.unit
def test_perform_train_pipeline_calls_get_model_output_path(
    reranker: Reranker,
    mock_get_model_output_path: MagicMock,
    mock_training_args: MagicMock,
    mock_silent_trainer: MagicMock,
    mock_os_path_exists: MagicMock,
    mock_os_remove: MagicMock
) -> None:
    """
    Test that _perform_train_pipeline calls get_model_output_path.
    
    Args:
        reranker (Reranker): The Reranker instance.
        mock_get_model_output_path (MagicMock): Mocked get_model_output_path function.
        mock_training_args (MagicMock): Mocked TrainingArguments class.
        mock_silent_trainer (MagicMock): Mocked SilentTrainer class.
        mock_os_path_exists (MagicMock): Mocked os.path.exists function.
        mock_os_remove (MagicMock): Mocked os.remove function.
    """
    
    user_instructions = ParsedModelInstructions(
        user_instructions=["healthcare"],
        language="english"
    )
    
    reranker._perform_train_pipeline(
        user_instructions=user_instructions,
        output_path="/custom/path"
    )
    
    mock_get_model_output_path.assert_called_once_with("/custom/path")


@pytest.mark.unit
def test_perform_train_pipeline_creates_training_args_with_correct_params(
    reranker: Reranker,
    mock_get_model_output_path: MagicMock,
    mock_training_args: MagicMock,
    mock_silent_trainer: MagicMock,
    mock_os_path_exists: MagicMock,
    mock_os_remove: MagicMock
) -> None:
    """
    Test that _perform_train_pipeline creates TrainingArguments with correct parameters.
    
    Args:
        reranker (Reranker): The Reranker instance.
        mock_get_model_output_path (MagicMock): Mocked get_model_output_path function.
        mock_training_args (MagicMock): Mocked TrainingArguments class.
        mock_silent_trainer (MagicMock): Mocked SilentTrainer class.
        mock_os_path_exists (MagicMock): Mocked os.path.exists function.
        mock_os_remove (MagicMock): Mocked os.remove function.
    """
    
    user_instructions = ParsedModelInstructions(
        user_instructions=["healthcare"],
        language="english"
    )
    
    reranker._perform_train_pipeline(
        user_instructions=user_instructions,
        output_path="/output",
        num_epochs=10
    )
    
    call_kwargs = mock_training_args.call_args[1]
    assert call_kwargs['output_dir'] == "/test/output/model"
    assert call_kwargs['num_train_epochs'] == 10
    assert call_kwargs['per_device_train_batch_size'] == 16
    assert call_kwargs['per_device_eval_batch_size'] == 16
    assert call_kwargs['save_strategy'] == "no"
    assert call_kwargs['logging_strategy'] == "no"
    assert call_kwargs['disable_tqdm'] is True
    assert call_kwargs['save_safetensors'] is True


@pytest.mark.unit
def test_perform_train_pipeline_creates_silent_trainer(
    reranker: Reranker,
    mock_get_model_output_path: MagicMock,
    mock_training_args: MagicMock,
    mock_silent_trainer: MagicMock,
    mock_os_path_exists: MagicMock,
    mock_os_remove: MagicMock
) -> None:
    """
    Test that _perform_train_pipeline creates SilentTrainer with correct parameters.
    
    Args:
        reranker (Reranker): The Reranker instance.
        mock_get_model_output_path (MagicMock): Mocked get_model_output_path function.
        mock_training_args (MagicMock): Mocked TrainingArguments class.
        mock_silent_trainer (MagicMock): Mocked SilentTrainer class.
        mock_os_path_exists (MagicMock): Mocked os.path.exists function.
        mock_os_remove (MagicMock): Mocked os.remove function.
    """
    
    user_instructions = ParsedModelInstructions(
        user_instructions=["healthcare"],
        language="english"
    )
    
    reranker._perform_train_pipeline(
        user_instructions=user_instructions,
        output_path="/output"
    )
    
    call_kwargs = mock_silent_trainer.call_args[1]
    assert call_kwargs['model'] == reranker._model
    assert 'train_dataset' in call_kwargs
    assert 'eval_dataset' in call_kwargs
    assert 'callbacks' in call_kwargs


@pytest.mark.unit
def test_perform_train_pipeline_calls_trainer_train(
    reranker: Reranker,
    mock_get_model_output_path: MagicMock,
    mock_training_args: MagicMock,
    mock_silent_trainer: MagicMock,
    mock_os_path_exists: MagicMock,
    mock_os_remove: MagicMock
) -> None:
    """
    Test that _perform_train_pipeline calls trainer.train().
    
    Args:
        reranker (Reranker): The Reranker instance.
        mock_get_model_output_path (MagicMock): Mocked get_model_output_path function.
        mock_training_args (MagicMock): Mocked TrainingArguments class.
        mock_silent_trainer (MagicMock): Mocked SilentTrainer class.
        mock_os_path_exists (MagicMock): Mocked os.path.exists function.
        mock_os_remove (MagicMock): Mocked os.remove function.
    """
    
    user_instructions = ParsedModelInstructions(
        user_instructions=["healthcare"],
        language="english"
    )
    
    reranker._perform_train_pipeline(
        user_instructions=user_instructions,
        output_path="/output"
    )
    
    trainer_instance = mock_silent_trainer.return_value
    trainer_instance.train.assert_called_once()


@pytest.mark.unit
def test_perform_train_pipeline_calls_save_model(
    reranker: Reranker,
    mock_get_model_output_path: MagicMock,
    mock_training_args: MagicMock,
    mock_silent_trainer: MagicMock,
    mock_os_path_exists: MagicMock,
    mock_os_remove: MagicMock
) -> None:
    """
    Test that _perform_train_pipeline calls trainer.save_model().
    
    Args:
        reranker (Reranker): The Reranker instance.
        mock_get_model_output_path (MagicMock): Mocked get_model_output_path function.
        mock_training_args (MagicMock): Mocked TrainingArguments class.
        mock_silent_trainer (MagicMock): Mocked SilentTrainer class.
        mock_os_path_exists (MagicMock): Mocked os.path.exists function.
        mock_os_remove (MagicMock): Mocked os.remove function.
    """
    
    user_instructions = ParsedModelInstructions(
        user_instructions=["healthcare"],
        language="english"
    )
    
    reranker._perform_train_pipeline(
        user_instructions=user_instructions,
        output_path="/output"
    )
    
    trainer_instance = mock_silent_trainer.return_value
    trainer_instance.save_model.assert_called_once()


@pytest.mark.unit
def test_perform_train_pipeline_removes_training_args_file(
    reranker: Reranker,
    mock_get_model_output_path: MagicMock,
    mock_training_args: MagicMock,
    mock_silent_trainer: MagicMock,
    mock_os_path_exists: MagicMock,
    mock_os_remove: MagicMock
) -> None:
    """
    Test that _perform_train_pipeline removes training_args.bin file if it exists.
    
    Args:
        reranker (Reranker): The Reranker instance.
        mock_get_model_output_path (MagicMock): Mocked get_model_output_path function.
        mock_training_args (MagicMock): Mocked TrainingArguments class.
        mock_silent_trainer (MagicMock): Mocked SilentTrainer class.
        mock_os_path_exists (MagicMock): Mocked os.path.exists function.
        mock_os_remove (MagicMock): Mocked os.remove function.
    """
    
    user_instructions = ParsedModelInstructions(
        user_instructions=["healthcare"],
        language="english"
    )
    
    reranker._perform_train_pipeline(
        user_instructions=user_instructions,
        output_path="/output"
    )
    
    mock_os_path_exists.assert_called_once_with("/test/output/model/training_args.bin")
    mock_os_remove.assert_called_once_with("/test/output/model/training_args.bin")


@pytest.mark.unit
def test_perform_train_pipeline_returns_train_output(
    reranker: Reranker,
    mock_get_model_output_path: MagicMock,
    mock_training_args: MagicMock,
    mock_silent_trainer: MagicMock,
    mock_os_path_exists: MagicMock,
    mock_os_remove: MagicMock
) -> None:
    """
    Test that _perform_train_pipeline returns TrainOutput from trainer.
    
    Args:
        reranker (Reranker): The Reranker instance.
        mock_get_model_output_path (MagicMock): Mocked get_model_output_path function.
        mock_training_args (MagicMock): Mocked TrainingArguments class.
        mock_silent_trainer (MagicMock): Mocked SilentTrainer class.
        mock_os_path_exists (MagicMock): Mocked os.path.exists function.
        mock_os_remove (MagicMock): Mocked os.remove function.
    """
    
    user_instructions = ParsedModelInstructions(
        user_instructions=["healthcare"],
        language="english"
    )
    
    result = reranker._perform_train_pipeline(
        user_instructions=user_instructions,
        output_path="/output"
    )
    
    assert isinstance(result, TrainOutput)
    assert result.global_step == 100
    assert result.training_loss == 0.5


@pytest.mark.unit
def test_perform_train_pipeline_uses_default_num_samples(
    reranker: Reranker,
    mock_get_model_output_path: MagicMock,
    mock_training_args: MagicMock,
    mock_silent_trainer: MagicMock,
    mock_os_path_exists: MagicMock,
    mock_os_remove: MagicMock,
    mocker: MockerFixture
) -> None:
    """
    Test that _perform_train_pipeline uses default num_samples when not provided.
    
    Args:
        reranker (Reranker): The Reranker instance.
        mock_get_model_output_path (MagicMock): Mocked get_model_output_path function.
        mock_training_args (MagicMock): Mocked TrainingArguments class.
        mock_silent_trainer (MagicMock): Mocked SilentTrainer class.
        mock_os_path_exists (MagicMock): Mocked os.path.exists function.
        mock_os_remove (MagicMock): Mocked os.remove function.
        mocker (MockerFixture): The pytest-mock fixture for mocking.
    """
    
    mock_build = mocker.spy(reranker, '_build_tokenized_train_ds')
    
    user_instructions = ParsedModelInstructions(
        user_instructions=["healthcare"],
        language="english"
    )
    
    reranker._perform_train_pipeline(
        user_instructions=user_instructions,
        output_path="/output"
    )
    
    call_kwargs = mock_build.call_args[1]
    assert call_kwargs['num_samples'] == config.DEFAULT_SYNTHEX_DATAPOINT_NUM


@pytest.mark.unit
def test_perform_train_pipeline_uses_default_num_epochs(
    reranker: Reranker,
    mock_get_model_output_path: MagicMock,
    mock_training_args: MagicMock,
    mock_silent_trainer: MagicMock,
    mock_os_path_exists: MagicMock,
    mock_os_remove: MagicMock
) -> None:
    """
    Test that _perform_train_pipeline uses default num_epochs when not provided.
    
    Args:
        reranker (Reranker): The Reranker instance.
        mock_get_model_output_path (MagicMock): Mocked get_model_output_path function.
        mock_training_args (MagicMock): Mocked TrainingArguments class.
        mock_silent_trainer (MagicMock): Mocked SilentTrainer class.
        mock_os_path_exists (MagicMock): Mocked os.path.exists function.
        mock_os_remove (MagicMock): Mocked os.remove function.
    """
    
    user_instructions = ParsedModelInstructions(
        user_instructions=["healthcare"],
        language="english"
    )
    
    reranker._perform_train_pipeline(
        user_instructions=user_instructions,
        output_path="/output"
    )
    
    call_kwargs = mock_training_args.call_args[1]
    assert call_kwargs['num_train_epochs'] == 3


@pytest.mark.unit
def test_perform_train_pipeline_does_not_remove_file_if_not_exists(
    reranker: Reranker,
    mock_get_model_output_path: MagicMock,
    mock_training_args: MagicMock,
    mock_silent_trainer: MagicMock,
    mock_os_remove: MagicMock,
    mocker: MockerFixture
) -> None:
    """
    Test that _perform_train_pipeline doesn't call os.remove if file doesn't exist.
    
    Args:
        reranker (Reranker): The Reranker instance.
        mock_get_model_output_path (MagicMock): Mocked get_model_output_path function.
        mock_training_args (MagicMock): Mocked TrainingArguments class.
        mock_silent_trainer (MagicMock): Mocked SilentTrainer class.
        mock_os_remove (MagicMock): Mocked os.remove function.
        mocker (MockerFixture): The pytest-mock fixture for mocking.
    """
    
    mocker.patch('artifex.models.reranker.reranker.os.path.exists', return_value=False)
    
    user_instructions = ParsedModelInstructions(
        user_instructions=["healthcare"],
        language="english"
    )
    
    reranker._perform_train_pipeline(
        user_instructions=user_instructions,
        output_path="/output"
    )
    
    mock_os_remove.assert_not_called()
    
    
@pytest.mark.unit
def test_perform_train_pipeline_calls_should_disable_cuda_with_device(
    reranker: Reranker,
    mock_get_model_output_path: MagicMock,
    mock_training_args: MagicMock,
    mock_silent_trainer: MagicMock,
    mock_os_path_exists: MagicMock,
    mock_os_remove: MagicMock,
    mocker: MockerFixture
) -> None:
    """
    Test that _perform_train_pipeline calls _should_disable_cuda with device parameter.
    
    Args:
        reranker (Reranker): The Reranker instance.
        mock_get_model_output_path (MagicMock): Mocked get_model_output_path function.
        mock_training_args (MagicMock): Mocked TrainingArguments class.
        mock_silent_trainer (MagicMock): Mocked SilentTrainer class.
        mock_os_path_exists (MagicMock): Mocked os.path.exists function.
        mock_os_remove (MagicMock): Mocked os.remove function.
        mocker (MockerFixture): The pytest-mock fixture for mocking.
    """
    
    mock_should_disable = mocker.spy(reranker, '_should_disable_cuda')
    
    user_instructions = ParsedModelInstructions(
        user_instructions=["healthcare"],
        language="english"
    )
    
    reranker._perform_train_pipeline(
        user_instructions=user_instructions,
        output_path="/output",
        num_samples=200,
        num_epochs=5,
        device=0
    )
    
    mock_should_disable.assert_called_once_with(0)


@pytest.mark.unit
def test_perform_train_pipeline_sets_use_cpu_false_when_device_is_0(
    reranker: Reranker,
    mock_get_model_output_path: MagicMock,
    mock_training_args: MagicMock,
    mock_silent_trainer: MagicMock,
    mock_os_path_exists: MagicMock,
    mock_os_remove: MagicMock
) -> None:
    """
    Test that _perform_train_pipeline sets use_cpu=False when device is 0 (GPU).
    
    Args:
        reranker (Reranker): The Reranker instance.
        mock_get_model_output_path (MagicMock): Mocked get_model_output_path function.
        mock_training_args (MagicMock): Mocked TrainingArguments class.
        mock_silent_trainer (MagicMock): Mocked SilentTrainer class.
        mock_os_path_exists (MagicMock): Mocked os.path.exists function.
        mock_os_remove (MagicMock): Mocked os.remove function.
    """
    
    user_instructions = ParsedModelInstructions(
        user_instructions=["healthcare"],
        language="english"
    )
    
    reranker._perform_train_pipeline(
        user_instructions=user_instructions,
        output_path="/output",
        num_samples=100,
        num_epochs=3,
        device=0
    )
    
    call_kwargs = mock_training_args.call_args[1]
    assert call_kwargs['use_cpu'] is False


@pytest.mark.unit
def test_perform_train_pipeline_sets_use_cpu_true_when_device_is_minus_1(
    reranker: Reranker,
    mock_get_model_output_path: MagicMock,
    mock_training_args: MagicMock,
    mock_silent_trainer: MagicMock,
    mock_os_path_exists: MagicMock,
    mock_os_remove: MagicMock
) -> None:
    """
    Test that _perform_train_pipeline sets use_cpu=True when device is -1 (CPU/MPS).
    
    Args:
        reranker (Reranker): The Reranker instance.
        mock_get_model_output_path (MagicMock): Mocked get_model_output_path function.
        mock_training_args (MagicMock): Mocked TrainingArguments class.
        mock_silent_trainer (MagicMock): Mocked SilentTrainer class.
        mock_os_path_exists (MagicMock): Mocked os.path.exists function.
        mock_os_remove (MagicMock): Mocked os.remove function.
    """
    
    user_instructions = ParsedModelInstructions(
        user_instructions=["healthcare"],
        language="english"
    )
    
    reranker._perform_train_pipeline(
        user_instructions=user_instructions,
        output_path="/output",
        num_samples=100,
        num_epochs=3,
        device=-1
    )
    
    call_kwargs = mock_training_args.call_args[1]
    assert call_kwargs['use_cpu'] is True


@pytest.mark.unit
def test_perform_train_pipeline_with_device_none_uses_default(
    reranker: Reranker,
    mock_get_model_output_path: MagicMock,
    mock_training_args: MagicMock,
    mock_silent_trainer: MagicMock,
    mock_os_path_exists: MagicMock,
    mock_os_remove: MagicMock,
    mocker: MockerFixture
) -> None:
    """
    Test that _perform_train_pipeline handles device=None correctly.
    
    Args:
        reranker (Reranker): The Reranker instance.
        mock_get_model_output_path (MagicMock): Mocked get_model_output_path function.
        mock_training_args (MagicMock): Mocked TrainingArguments class.
        mock_silent_trainer (MagicMock): Mocked SilentTrainer class.
        mock_os_path_exists (MagicMock): Mocked os.path.exists function.
        mock_os_remove (MagicMock): Mocked os.remove function.
        mocker (MockerFixture): The pytest-mock fixture for mocking.
    """
    
    mock_should_disable = mocker.spy(reranker, '_should_disable_cuda')
    
    user_instructions = ParsedModelInstructions(
        user_instructions=["healthcare"],
        language="english"
    )
    
    reranker._perform_train_pipeline(
        user_instructions=user_instructions,
        output_path="/output",
        num_samples=100,
        num_epochs=3,
        device=None
    )
    
    mock_should_disable.assert_called_once_with(None)


import pytest
from pytest_mock import MockerFixture
from synthex import Synthex
from unittest.mock import MagicMock

from artifex.models.reranker import Reranker
from artifex.core import ParsedModelInstructions
from artifex.config import config


@pytest.fixture
def mock_dependencies(mocker: MockerFixture) -> None:
    """
    Fixture to mock external dependencies for Reranker.
    
    Args:
        mocker (MockerFixture): The pytest-mock fixture for mocking.
    """
    
    mocker.patch(
        'artifex.models.reranker.reranker.AutoModelForSequenceClassification.from_pretrained',
        return_value=MagicMock()
    )
    mocker.patch(
        'artifex.models.reranker.reranker.AutoTokenizer.from_pretrained',
        return_value=MagicMock()
    )


@pytest.fixture
def mock_synthex(mocker: MockerFixture) -> Synthex:
    """
    Fixture to create a mock Synthex instance.
    
    Args:
        mocker (MockerFixture): The pytest-mock fixture for mocking.
    
    Returns:
        Synthex: A mocked Synthex instance.
    """
    
    return mocker.MagicMock()


@pytest.fixture
def reranker(mock_dependencies: None, mock_synthex: Synthex) -> Reranker:
    """
    Fixture to create a Reranker instance for testing.
    
    Args:
        mock_dependencies (None): Fixture that mocks external dependencies.
        mock_synthex (Synthex): A mocked Synthex instance.
    
    Returns:
        Reranker: A Reranker instance.
    """
    
    return Reranker(synthex=mock_synthex)


@pytest.mark.unit
def test_get_data_gen_instr_returns_list_of_strings(
    reranker: Reranker
) -> None:
    """
    Test that _get_data_gen_instr returns a list of strings.
    
    Args:
        reranker (Reranker): The Reranker instance.
    """
    
    user_instr = ParsedModelInstructions(
        user_instructions=["scientific research"],
        language="english"
    )
    
    result = reranker._get_data_gen_instr(user_instr)
    
    assert isinstance(result, list)
    assert all(isinstance(item, str) for item in result)


@pytest.mark.unit
def test_get_data_gen_instr_formats_language_placeholder(
    reranker: Reranker
) -> None:
    """
    Test that _get_data_gen_instr correctly formats the language placeholder.
    
    Args:
        reranker (Reranker): The Reranker instance.
    """
    
    user_instr = ParsedModelInstructions(
        user_instructions=["healthcare"],
        language="spanish"
    )
    
    result = reranker._get_data_gen_instr(user_instr)
    
    # Find instruction that contains language placeholder
    language_instructions = [instr for instr in result if "spanish" in instr.lower()]
    assert len(language_instructions) > 0


@pytest.mark.unit
def test_get_data_gen_instr_formats_domain_placeholder(
    reranker: Reranker
) -> None:
    """
    Test that _get_data_gen_instr correctly formats the domain placeholder.
    
    Args:
        reranker (Reranker): The Reranker instance.
    """
    
    user_instr = ParsedModelInstructions(
        user_instructions=["scientific research"],
        language="english"
    )
    
    result = reranker._get_data_gen_instr(user_instr)
    
    # Find instruction that contains domain
    domain_instructions = [instr for instr in result if "scientific research" in instr]
    assert len(domain_instructions) > 0


@pytest.mark.unit
def test_get_data_gen_instr_returns_correct_number_of_instructions(
    reranker: Reranker
) -> None:
    """
    Test that _get_data_gen_instr returns same number of instructions as system instructions.
    
    Args:
        reranker (Reranker): The Reranker instance.
    """
    
    user_instr = ParsedModelInstructions(
        user_instructions=["medical research"],
        language="english"
    )
    
    result = reranker._get_data_gen_instr(user_instr)
    
    assert len(result) == len(reranker._system_data_gen_instr_val)


@pytest.mark.unit
def test_get_data_gen_instr_with_single_domain(
    reranker: Reranker
) -> None:
    """
    Test _get_data_gen_instr with single domain item.
    
    Args:
        reranker (Reranker): The Reranker instance.
    """
    
    user_instr = ParsedModelInstructions(
        user_instructions=["technology"],
        language="english"
    )
    
    result = reranker._get_data_gen_instr(user_instr)
    
    assert any("technology" in instr for instr in result)


@pytest.mark.unit
def test_get_data_gen_instr_with_different_languages(
    reranker: Reranker
) -> None:
    """
    Test _get_data_gen_instr with different language values.
    
    Args:
        reranker (Reranker): The Reranker instance.
    """
    
    for language in ["english", "spanish", "french", "german", "chinese"]:
        user_instr = ParsedModelInstructions(
            user_instructions=["healthcare"],
            language=language
        )
        
        result = reranker._get_data_gen_instr(user_instr)
        
        assert any(language in instr for instr in result)


@pytest.mark.unit
def test_get_data_gen_instr_with_special_characters_in_domain(
    reranker: Reranker
) -> None:
    """
    Test _get_data_gen_instr with special characters in domain.
    
    Args:
        reranker (Reranker): The Reranker instance.
    """
    
    user_instr = ParsedModelInstructions(
        user_instructions=["research & development!@#$%"],
        language="english"
    )
    
    result = reranker._get_data_gen_instr(user_instr)
    
    result_str = " ".join(result)
    assert "research & development!@#$%" in result_str


@pytest.mark.unit
def test_get_data_gen_instr_with_unicode_in_domain(
    reranker: Reranker
) -> None:
    """
    Test _get_data_gen_instr with unicode characters in domain.
    
    Args:
        reranker (Reranker): The Reranker instance.
    """
    
    user_instr = ParsedModelInstructions(
        user_instructions=["investigación científica 你好"],
        language="spanish"
    )
    
    result = reranker._get_data_gen_instr(user_instr)
    
    result_str = " ".join(result)
    assert "investigación científica 你好" in result_str or "spanish" in result_str


@pytest.mark.unit
def test_get_data_gen_instr_with_unicode_language(
    reranker: Reranker
) -> None:
    """
    Test _get_data_gen_instr with unicode language parameter.
    
    Args:
        reranker (Reranker): The Reranker instance.
    """
    
    user_instr = ParsedModelInstructions(
        user_instructions=["healthcare"],
        language="中文"
    )
    
    result = reranker._get_data_gen_instr(user_instr)
    
    assert any("中文" in instr for instr in result)


@pytest.mark.unit
def test_get_data_gen_instr_with_long_domain_string(
    reranker: Reranker
) -> None:
    """
    Test _get_data_gen_instr with long domain string.
    
    Args:
        reranker (Reranker): The Reranker instance.
    """
    
    long_domain = "This is a very long description of a domain that spans multiple sentences and provides detailed information about the area of specialization."
    
    user_instr = ParsedModelInstructions(
        user_instructions=[long_domain],
        language="english"
    )
    
    result = reranker._get_data_gen_instr(user_instr)
    
    assert any(long_domain in instr for instr in result)


@pytest.mark.unit
def test_get_data_gen_instr_does_not_modify_input(
    reranker: Reranker
) -> None:
    """
    Test that _get_data_gen_instr does not modify the input ParsedModelInstructions.
    
    Args:
        reranker (Reranker): The Reranker instance.
    """
    
    user_instr = ParsedModelInstructions(
        user_instructions=["healthcare"],
        language="english"
    )
    
    original_instructions = user_instr.user_instructions.copy()
    original_language = user_instr.language
    
    reranker._get_data_gen_instr(user_instr)
    
    assert user_instr.user_instructions == original_instructions
    assert user_instr.language == original_language


@pytest.mark.unit
def test_get_data_gen_instr_with_whitespace_in_domain(
    reranker: Reranker
) -> None:
    """
    Test _get_data_gen_instr with whitespace in domain.
    
    Args:
        reranker (Reranker): The Reranker instance.
    """
    
    user_instr = ParsedModelInstructions(
        user_instructions=["  healthcare  "],
        language="english"
    )
    
    result = reranker._get_data_gen_instr(user_instr)
    
    result_str = " ".join(result)
    # Whitespace should be preserved
    assert "  healthcare  " in result_str or "healthcare" in result_str


@pytest.mark.unit
def test_get_data_gen_instr_with_quotes_in_domain(
    reranker: Reranker
) -> None:
    """
    Test _get_data_gen_instr with quotes in domain.
    
    Args:
        reranker (Reranker): The Reranker instance.
    """
    
    user_instr = ParsedModelInstructions(
        user_instructions=['domain with "double quotes"'],
        language="english"
    )
    
    result = reranker._get_data_gen_instr(user_instr)
    
    result_str = " ".join(result)
    assert '"double quotes"' in result_str


@pytest.mark.unit
def test_get_data_gen_instr_with_newlines_in_domain(
    reranker: Reranker
) -> None:
    """
    Test _get_data_gen_instr with newlines in domain.
    
    Args:
        reranker (Reranker): The Reranker instance.
    """
    
    user_instr = ParsedModelInstructions(
        user_instructions=["domain\nwith\nnewlines"],
        language="english"
    )
    
    result = reranker._get_data_gen_instr(user_instr)
    
    assert any("domain\\nwith\\nnewlines" in instr for instr in result)


@pytest.mark.unit
def test_get_data_gen_instr_with_numeric_domain(
    reranker: Reranker
) -> None:
    """
    Test _get_data_gen_instr with numeric strings in domain.
    
    Args:
        reranker (Reranker): The Reranker instance.
    """
    
    user_instr = ParsedModelInstructions(
        user_instructions=["123 456.789"],
        language="english"
    )
    
    result = reranker._get_data_gen_instr(user_instr)
    
    result_str = " ".join(result)
    assert "123 456.789" in result_str


@pytest.mark.unit
def test_get_data_gen_instr_formats_all_system_instructions(
    reranker: Reranker
) -> None:
    """
    Test that _get_data_gen_instr formats all system instructions.
    
    Args:
        reranker (Reranker): The Reranker instance.
    """
    
    user_instr = ParsedModelInstructions(
        user_instructions=["healthcare"],
        language="english"
    )
    
    result = reranker._get_data_gen_instr(user_instr)
    
    # Result should have same length as system instructions
    assert len(result) == len(reranker._system_data_gen_instr_val)
    
    # All should be non-empty strings
    assert all(len(instr) > 0 for instr in result)


@pytest.mark.unit
def test_get_data_gen_instr_with_domain_none(
    reranker: Reranker
) -> None:
    """
    Test _get_data_gen_instr when domain is None in ParsedModelInstructions.
    
    Args:
        reranker (Reranker): The Reranker instance.
    """
    
    user_instr = ParsedModelInstructions(
        user_instructions=["healthcare"],
        language="english",
        domain=None
    )
    
    result = reranker._get_data_gen_instr(user_instr)
    
    # Should work even without explicit domain field
    assert isinstance(result, list)
    assert len(result) > 0


@pytest.mark.unit
def test_get_data_gen_instr_with_mixed_case_language(
    reranker: Reranker
) -> None:
    """
    Test _get_data_gen_instr with mixed case language parameter.
    
    Args:
        reranker (Reranker): The Reranker instance.
    """
    
    user_instr = ParsedModelInstructions(
        user_instructions=["healthcare"],
        language="EnGLisH"
    )
    
    result = reranker._get_data_gen_instr(user_instr)
    
    assert any("EnGLisH" in instr for instr in result)


@pytest.mark.unit
def test_get_data_gen_instr_with_complex_punctuation(
    reranker: Reranker
) -> None:
    """
    Test _get_data_gen_instr with complex punctuation in domain.
    
    Args:
        reranker (Reranker): The Reranker instance.
    """
    
    user_instr = ParsedModelInstructions(
        user_instructions=["domain (with [nested {punctuation}])"],
        language="english"
    )
    
    result = reranker._get_data_gen_instr(user_instr)
    
    assert any("domain (with [nested {punctuation}])" in instr for instr in result)


@pytest.mark.unit
def test_get_data_gen_instr_with_emoji_in_domain(
    reranker: Reranker
) -> None:
    """
    Test _get_data_gen_instr with emoji characters in domain.
    
    Args:
        reranker (Reranker): The Reranker instance.
    """
    
    user_instr = ParsedModelInstructions(
        user_instructions=["healthcare 🏥💊"],
        language="english"
    )
    
    result = reranker._get_data_gen_instr(user_instr)
    
    assert any("healthcare 🏥💊" in instr for instr in result)


@pytest.mark.unit
def test_get_data_gen_instr_result_contains_only_strings(
    reranker: Reranker
) -> None:
    """
    Test that _get_data_gen_instr result contains only string types.
    
    Args:
        reranker (Reranker): The Reranker instance.
    """
    
    user_instr = ParsedModelInstructions(
        user_instructions=["healthcare"],
        language="english"
    )
    
    result = reranker._get_data_gen_instr(user_instr)
    
    for item in result:
        assert isinstance(item, str)
        assert not isinstance(item, (list, dict, tuple))


@pytest.mark.unit
def test_get_data_gen_instr_with_backslashes(
    reranker: Reranker
) -> None:
    """
    Test _get_data_gen_instr with backslashes in domain.
    
    Args:
        reranker (Reranker): The Reranker instance.
    """
    
    user_instr = ParsedModelInstructions(
        user_instructions=["domain\\with\\backslashes"],
        language="english"
    )
    
    result = reranker._get_data_gen_instr(user_instr)
    
    assert any("domain\\\\with\\\\backslashes" in instr for instr in result)


@pytest.mark.unit
def test_get_data_gen_instr_consecutive_calls_produce_same_result(
    reranker: Reranker
) -> None:
    """
    Test that consecutive calls to _get_data_gen_instr with same input produce same result.
    
    Args:
        reranker (Reranker): The Reranker instance.
    """
    
    user_instr = ParsedModelInstructions(
        user_instructions=["healthcare"],
        language="english"
    )
    
    result1 = reranker._get_data_gen_instr(user_instr)
    result2 = reranker._get_data_gen_instr(user_instr)
    
    assert result1 == result2


@pytest.mark.unit
def test_get_data_gen_instr_with_empty_domain_list(
    reranker: Reranker
) -> None:
    """
    Test _get_data_gen_instr with empty user_instructions list.
    
    Args:
        reranker (Reranker): The Reranker instance.
    """
    
    user_instr = ParsedModelInstructions(
        user_instructions=[],
        language="english"
    )
    
    result = reranker._get_data_gen_instr(user_instr)
    
    # Should still return formatted system instructions
    assert len(result) == len(reranker._system_data_gen_instr_val)
    assert all(isinstance(item, str) for item in result)


@pytest.mark.unit
def test_get_data_gen_instr_domain_in_user_instructions_field(
    reranker: Reranker
) -> None:
    """
    Test that _get_data_gen_instr uses domain from user_instructions field.
    
    Args:
        reranker (Reranker): The Reranker instance.
    """
    
    user_instr = ParsedModelInstructions(
        user_instructions=["medical research"],
        language="english",
        domain="ignored_domain"
    )
    
    result = reranker._get_data_gen_instr(user_instr)
    
    # Should use user_instructions, not domain field
    assert any("medical research" in instr for instr in result)


@pytest.mark.unit
def test_get_data_gen_instr_with_tabs_in_domain(
    reranker: Reranker
) -> None:
    """
    Test _get_data_gen_instr with tab characters in domain.
    
    Args:
        reranker (Reranker): The Reranker instance.
    """
    
    user_instr = ParsedModelInstructions(
        user_instructions=["domain\twith\ttabs"],
        language="english"
    )
    
    result = reranker._get_data_gen_instr(user_instr)
    
    assert any("domain\\twith\\ttabs" in instr for instr in result)


@pytest.mark.unit
def test_get_data_gen_instr_with_multiple_domain_items(
    reranker: Reranker
) -> None:
    """
    Test _get_data_gen_instr when user_instructions contains multiple items (edge case).
    
    Args:
        reranker (Reranker): The Reranker instance.
    """
    
    user_instr = ParsedModelInstructions(
        user_instructions=["domain1", "domain2"],
        language="english"
    )
    
    result = reranker._get_data_gen_instr(user_instr)
    
    # Should format with the list as is
    assert isinstance(result, list)
    assert len(result) == len(reranker._system_data_gen_instr_val)


import pytest
from pytest_mock import MockerFixture
from synthex import Synthex

from artifex.models import ClassificationModel
from artifex.core import ClassificationResponse


@pytest.fixture
def mock_synthex(mocker: MockerFixture) -> Synthex:
    """
    Fixture to create a mock Synthex instance.
    Args:
        mocker (MockerFixture): The pytest-mock fixture for mocking.
    Returns:
        Synthex: A mocked Synthex instance.
    """
    
    return mocker.MagicMock()


@pytest.fixture
def mock_pipeline(mocker: MockerFixture) -> MockerFixture:
    """
    Fixture to mock transformers.pipeline.
    Args:
        mocker (MockerFixture): The pytest-mock fixture for mocking.
    Returns:
        MockerFixture: Mocked pipeline function.
    """
    
    mock = mocker.patch("artifex.models.classification.classification_model.pipeline")
    return mock


@pytest.fixture
def concrete_model(mock_synthex: Synthex, mocker: MockerFixture) -> ClassificationModel:
    """
    Fixture to create a concrete ClassificationModel instance for testing.
    Args:
        mock_synthex (Synthex): A mocked Synthex instance.
        mocker (MockerFixture): The pytest-mock fixture for mocking.
    Returns:
        ClassificationModel: A concrete implementation of ClassificationModel.
    """
    
    from synthex.models import JobOutputSchemaDefinition
    from datasets import ClassLabel
    from transformers.trainer_utils import TrainOutput
    
    # Mock the transformers components
    mocker.patch(
        "transformers.AutoModelForSequenceClassification.from_pretrained",
        return_value=mocker.MagicMock()
    )
    mocker.patch(
        "transformers.AutoTokenizer.from_pretrained",
        return_value=mocker.MagicMock()
    )
    
    class ConcreteClassificationModel(ClassificationModel):
        """Concrete implementation of ClassificationModel for testing purposes."""
        
        @property
        def _base_model_name(self) -> str:
            return "distilbert-base-uncased"
        
        @property
        def _token_keys(self) -> list[str]:
            return ["text"]
        
        @property
        def _synthetic_data_schema(self) -> JobOutputSchemaDefinition:
            return JobOutputSchemaDefinition(
                text={"type": "string"},
                label={"type": "integer"}
            )
            
        @property
        def _system_data_gen_instr(self) -> list[str]:
            return ["system instruction 1", "system instruction 2"]
        
        @property
        def _labels(self) -> ClassLabel:
            return ClassLabel(names=["negative", "positive"])
        
        def _parse_user_instructions(self, user_instructions: list[str]) -> list[str]:
            return user_instructions
        
        def _get_data_gen_instr(self, user_instr: list[str]) -> list[str]:
            return user_instr
        
        def _post_process_synthetic_dataset(self, synthetic_dataset_path: str):
            pass
        
        def _load_model(self, model_path: str):
            """Mock implementation of _load_model."""
            pass
        
        def train(
            self, instructions: list[str], output_path: str | None = None,
            num_samples: int = 500, num_epochs: int = 3
        ) -> TrainOutput:
            """Mock implementation of train."""
            return TrainOutput(global_step=100, training_loss=0.5, metrics={})
    
    return ConcreteClassificationModel(mock_synthex)


@pytest.mark.unit
def test_call_with_single_string_returns_classification_responses(
    concrete_model: ClassificationModel, mock_pipeline: MockerFixture
):
    """
    Test that __call__ with a single string returns list of ClassificationResponse.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        mock_pipeline (MockerFixture): Mocked pipeline function.
    """

    mock_classifier = mock_pipeline.return_value
    mock_classifier.return_value = [{"label": "positive", "score": 0.95}]
    
    result = concrete_model("This is a test")
    
    assert len(result) == 1
    assert isinstance(result[0], ClassificationResponse)
    assert result[0].label == "positive"
    assert result[0].score == 0.95


@pytest.mark.unit
def test_call_creates_pipeline_with_correct_arguments(
    concrete_model: ClassificationModel, mock_pipeline: MockerFixture
):
    """
    Test that __call__ creates pipeline with correct model and tokenizer.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        mock_pipeline (MockerFixture): Mocked pipeline function.
    """

    mock_classifier = mock_pipeline.return_value
    mock_classifier.return_value = [{"label": "positive", "score": 0.95}]
    
    concrete_model("test text")
    
    mock_pipeline.assert_called_once_with(
        "text-classification",
        model=concrete_model._model,
        tokenizer=concrete_model._tokenizer,
        device=-1
    )


@pytest.mark.unit
def test_call_passes_text_to_classifier(
    concrete_model: ClassificationModel, mock_pipeline: MockerFixture
):
    """
    Test that __call__ passes the input text to the classifier.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        mock_pipeline (MockerFixture): Mocked pipeline function.
    """

    mock_classifier = mock_pipeline.return_value
    mock_classifier.return_value = [{"label": "positive", "score": 0.95}]
    
    input_text = "This is a test sentence"
    concrete_model(input_text)
    
    mock_classifier.assert_called_once_with(input_text)


@pytest.mark.unit
def test_call_with_list_of_strings(
    concrete_model: ClassificationModel, mock_pipeline: MockerFixture
):
    """
    Test that __call__ works with a list of strings.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        mock_pipeline (MockerFixture): Mocked pipeline function.
    """

    mock_classifier = mock_pipeline.return_value
    mock_classifier.return_value = [
        {"label": "positive", "score": 0.95},
        {"label": "negative", "score": 0.88}
    ]
    
    result = concrete_model(["text1", "text2"])
    
    assert len(result) == 2
    assert result[0].label == "positive"
    assert result[1].label == "negative"


@pytest.mark.unit
def test_call_returns_empty_list_when_classifier_returns_empty(
    concrete_model: ClassificationModel, mock_pipeline: MockerFixture
):
    """
    Test that __call__ returns empty list when classifier returns empty results.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        mock_pipeline (MockerFixture): Mocked pipeline function.
    """

    mock_classifier = mock_pipeline.return_value
    mock_classifier.return_value = []
    
    result = concrete_model("test")
    
    assert result == []


@pytest.mark.unit
def test_call_returns_empty_list_when_classifier_returns_none(
    concrete_model: ClassificationModel, mock_pipeline: MockerFixture
):
    """
    Test that __call__ returns empty list when classifier returns None.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        mock_pipeline (MockerFixture): Mocked pipeline function.
    """

    mock_classifier = mock_pipeline.return_value
    mock_classifier.return_value = None
    
    result = concrete_model("test")
    
    assert result == []


@pytest.mark.unit
def test_call_with_multiple_classifications(
    concrete_model: ClassificationModel, mock_pipeline: MockerFixture
):
    """
    Test that __call__ handles multiple classification results.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        mock_pipeline (MockerFixture): Mocked pipeline function.
    """

    mock_classifier = mock_pipeline.return_value
    mock_classifier.return_value = [
        {"label": "positive", "score": 0.95},
        {"label": "negative", "score": 0.85},
        {"label": "neutral", "score": 0.75}
    ]
    
    result = concrete_model(["text1", "text2", "text3"])
    
    assert len(result) == 3
    assert all(isinstance(r, ClassificationResponse) for r in result)


@pytest.mark.unit
def test_call_preserves_classification_order(
    concrete_model: ClassificationModel, mock_pipeline: MockerFixture
):
    """
    Test that __call__ preserves the order of classifications.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        mock_pipeline (MockerFixture): Mocked pipeline function.
    """

    mock_classifier = mock_pipeline.return_value
    mock_classifier.return_value = [
        {"label": "label1", "score": 0.1},
        {"label": "label2", "score": 0.2},
        {"label": "label3", "score": 0.3}
    ]
    
    result = concrete_model(["text1", "text2", "text3"])
    
    assert result[0].label == "label1"
    assert result[0].score == 0.1
    assert result[1].label == "label2"
    assert result[1].score == 0.2
    assert result[2].label == "label3"
    assert result[2].score == 0.3


@pytest.mark.unit
def test_call_with_high_confidence_score(
    concrete_model: ClassificationModel, mock_pipeline: MockerFixture
):
    """
    Test that __call__ correctly handles high confidence scores.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        mock_pipeline (MockerFixture): Mocked pipeline function.
    """

    mock_classifier = mock_pipeline.return_value
    mock_classifier.return_value = [{"label": "positive", "score": 0.9999}]
    
    result = concrete_model("test")
    
    assert result[0].score == 0.9999


@pytest.mark.unit
def test_call_with_low_confidence_score(
    concrete_model: ClassificationModel, mock_pipeline: MockerFixture
):
    """
    Test that __call__ correctly handles low confidence scores.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        mock_pipeline (MockerFixture): Mocked pipeline function.
    """

    mock_classifier = mock_pipeline.return_value
    mock_classifier.return_value = [{"label": "negative", "score": 0.0001}]
    
    result = concrete_model("test")
    
    assert result[0].score == 0.0001


@pytest.mark.unit
def test_call_with_empty_string(
    concrete_model: ClassificationModel, mock_pipeline: MockerFixture
):
    """
    Test that __call__ handles empty string input.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        mock_pipeline (MockerFixture): Mocked pipeline function.
    """

    mock_classifier = mock_pipeline.return_value
    mock_classifier.return_value = [{"label": "neutral", "score": 0.5}]
    
    result = concrete_model("")
    
    mock_classifier.assert_called_once_with("")
    assert len(result) == 1


@pytest.mark.unit
def test_call_with_long_text(
    concrete_model: ClassificationModel, mock_pipeline: MockerFixture
):
    """
    Test that __call__ handles long text input.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        mock_pipeline (MockerFixture): Mocked pipeline function.
    """

    mock_classifier = mock_pipeline.return_value
    mock_classifier.return_value = [{"label": "positive", "score": 0.85}]
    
    long_text = "word " * 1000
    result = concrete_model(long_text)
    
    mock_classifier.assert_called_once_with(long_text)
    assert len(result) == 1


@pytest.mark.unit
def test_call_with_special_characters(
    concrete_model: ClassificationModel, mock_pipeline: MockerFixture
):
    """
    Test that __call__ handles text with special characters.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        mock_pipeline (MockerFixture): Mocked pipeline function.
    """

    mock_classifier = mock_pipeline.return_value
    mock_classifier.return_value = [{"label": "positive", "score": 0.9}]
    
    special_text = "Hello! @#$%^&*() 你好 🎉"
    result = concrete_model(special_text)
    
    mock_classifier.assert_called_once_with(special_text)
    assert result[0].label == "positive"


@pytest.mark.unit
def test_call_with_unicode_text(
    concrete_model: ClassificationModel, mock_pipeline: MockerFixture
):
    """
    Test that __call__ handles unicode text.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        mock_pipeline (MockerFixture): Mocked pipeline function.
    """

    mock_classifier = mock_pipeline.return_value
    mock_classifier.return_value = [{"label": "positive", "score": 0.92}]
    
    unicode_text = "Héllo Wörld 日本語 العربية"
    result = concrete_model(unicode_text)
    
    assert result[0].label == "positive"
    assert result[0].score == 0.92


@pytest.mark.unit
def test_call_creates_new_pipeline_each_time(
    concrete_model: ClassificationModel, mock_pipeline: MockerFixture
):
    """
    Test that __call__ creates a new pipeline for each invocation.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        mock_pipeline (MockerFixture): Mocked pipeline function.
    """

    mock_classifier = mock_pipeline.return_value
    mock_classifier.return_value = [{"label": "positive", "score": 0.9}]
    
    concrete_model("test1")
    concrete_model("test2")
    
    assert mock_pipeline.call_count == 2


@pytest.mark.unit
def test_call_with_whitespace_only(
    concrete_model: ClassificationModel, mock_pipeline: MockerFixture
):
    """
    Test that __call__ handles whitespace-only input.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        mock_pipeline (MockerFixture): Mocked pipeline function.
    """

    mock_classifier = mock_pipeline.return_value
    mock_classifier.return_value = [{"label": "neutral", "score": 0.5}]
    
    result = concrete_model("   \n\t  ")
    
    assert len(result) == 1


@pytest.mark.unit
def test_call_with_single_word(
    concrete_model: ClassificationModel, mock_pipeline: MockerFixture
):
    """
    Test that __call__ handles single word input.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        mock_pipeline (MockerFixture): Mocked pipeline function.
    """

    mock_classifier = mock_pipeline.return_value
    mock_classifier.return_value = [{"label": "positive", "score": 0.88}]
    
    result = concrete_model("excellent")
    
    assert result[0].label == "positive"
    assert result[0].score == 0.88


@pytest.mark.unit
def test_call_response_has_correct_attributes(
    concrete_model: ClassificationModel, mock_pipeline: MockerFixture
):
    """
    Test that ClassificationResponse objects have correct attributes.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        mock_pipeline (MockerFixture): Mocked pipeline function.
    """

    mock_classifier = mock_pipeline.return_value
    mock_classifier.return_value = [{"label": "test_label", "score": 0.777}]
    
    result = concrete_model("test")
    
    assert hasattr(result[0], "label")
    assert hasattr(result[0], "score")
    assert result[0].label == "test_label"
    assert result[0].score == 0.777


@pytest.mark.unit
def test_call_with_empty_list(
    concrete_model: ClassificationModel, mock_pipeline: MockerFixture
):
    """
    Test that __call__ handles empty list input.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        mock_pipeline (MockerFixture): Mocked pipeline function.
    """

    mock_classifier = mock_pipeline.return_value
    mock_classifier.return_value = []
    
    result = concrete_model([])
    
    mock_classifier.assert_called_once_with([])
    assert result == []


@pytest.mark.unit
def test_call_with_single_element_list(
    concrete_model: ClassificationModel, mock_pipeline: MockerFixture
):
    """
    Test that __call__ handles single element list.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        mock_pipeline (MockerFixture): Mocked pipeline function.
    """

    mock_classifier = mock_pipeline.return_value
    mock_classifier.return_value = [{"label": "positive", "score": 0.93}]
    
    result = concrete_model(["single text"])
    
    assert len(result) == 1
    assert result[0].label == "positive"


@pytest.mark.unit
def test_call_with_numeric_labels(
    concrete_model: ClassificationModel, mock_pipeline: MockerFixture
):
    """
    Test that __call__ handles numeric label correctly.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        mock_pipeline (MockerFixture): Mocked pipeline function.
    """

    mock_classifier = mock_pipeline.return_value
    mock_classifier.return_value = [{"label": "0", "score": 0.65}]
    
    result = concrete_model("test")
    
    assert result[0].label == "0"
    assert result[0].score == 0.65


@pytest.mark.unit
def test_call_with_label_prefix(
    concrete_model: ClassificationModel, mock_pipeline: MockerFixture
):
    """
    Test that __call__ handles label with LABEL_ prefix.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        mock_pipeline (MockerFixture): Mocked pipeline function.
    """

    mock_classifier = mock_pipeline.return_value
    mock_classifier.return_value = [{"label": "LABEL_0", "score": 0.82}]
    
    result = concrete_model("test")
    
    assert result[0].label == "LABEL_0"
    assert result[0].score == 0.82


@pytest.mark.unit
def test_call_with_exact_score_bounds(
    concrete_model: ClassificationModel, mock_pipeline: MockerFixture
):
    """
    Test that __call__ handles scores at exact boundaries (0.0 and 1.0).
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        mock_pipeline (MockerFixture): Mocked pipeline function.
    """

    mock_classifier = mock_pipeline.return_value
    mock_classifier.return_value = [
        {"label": "certain", "score": 1.0},
        {"label": "uncertain", "score": 0.0}
    ]
    
    result = concrete_model(["text1", "text2"])
    
    assert result[0].score == 1.0
    assert result[1].score == 0.0


@pytest.mark.unit
def test_call_list_comprehension_creates_all_responses(
    concrete_model: ClassificationModel, mock_pipeline: MockerFixture
):
    """
    Test that the list comprehension creates ClassificationResponse for all items.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        mock_pipeline (MockerFixture): Mocked pipeline function.
    """

    mock_classifier = mock_pipeline.return_value
    mock_classifier.return_value = [
        {"label": "a", "score": 0.1},
        {"label": "b", "score": 0.2},
        {"label": "c", "score": 0.3},
        {"label": "d", "score": 0.4},
        {"label": "e", "score": 0.5}
    ]
    
    result = concrete_model(["t1", "t2", "t3", "t4", "t5"])
    
    assert len(result) == 5
    assert all(isinstance(r, ClassificationResponse) for r in result)
    assert [r.label for r in result] == ["a", "b", "c", "d", "e"]
    

@pytest.mark.unit
def test_call_with_device_argument(
    concrete_model: ClassificationModel, mock_pipeline: MockerFixture
):
    """
    Test that __call__ correctly passes the device argument to the pipeline.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        mock_pipeline (MockerFixture): Mocked pipeline function.
    """

    mock_classifier = mock_pipeline.return_value
    mock_classifier.return_value = [{"label": "positive", "score": 0.9}]
    
    device = 0  # Example device ID
    concrete_model("test text", device=device)
    
    mock_pipeline.assert_called_once_with(
        "text-classification",
        model=concrete_model._model,
        tokenizer=concrete_model._tokenizer,
        device=device
    )
    

@pytest.mark.unit
def test_call_with_default_device(
    concrete_model: ClassificationModel, mock_pipeline: MockerFixture, mocker: MockerFixture
):
    """
    Test that __call__ calls _determine_default_device when device is None, and passes
    it to the pipeline.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        mock_pipeline (MockerFixture): Mocked pipeline function.
        mocker (MockerFixture): The pytest-mock fixture for mocking.
    """

    mock_classifier = mock_pipeline.return_value
    mock_classifier.return_value = [{"label": "positive", "score": 0.9}]
    
    # Mock the _determine_default_device method
    mock_device = 1
    mock_determine_device = mocker.patch.object(
        concrete_model, '_determine_default_device', return_value=mock_device
    )
    
    concrete_model("test text", device=None)
    
    mock_determine_device.assert_called_once()
    mock_pipeline.assert_called_once_with(
        "text-classification",
        model=concrete_model._model,
        tokenizer=concrete_model._tokenizer,
        device=mock_device
    )


@pytest.mark.unit
def test_call_logs_inference_with_decorator(
    concrete_model: ClassificationModel,
    mock_pipeline: MockerFixture,
    mocker: MockerFixture,
    tmp_path
):
    """
    Test that __call__ logs inference metrics through the @track_inference_calls decorator.
    
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        mock_pipeline (MockerFixture): Mocked pipeline function.
        mocker (MockerFixture): The pytest-mock fixture for mocking.
        tmp_path: Pytest fixture for temporary directory.
    """
    import json
    from pathlib import Path
    
    log_file = tmp_path / "inference.log"
    
    # Mock the config paths and decorator dependencies
    mocker.patch("artifex.core.decorators.logging.config.INFERENCE_LOGS_PATH", str(log_file))
    mocker.patch("artifex.core.decorators.logging._calculate_daily_inference_aggregates")
    
    # Mock psutil to avoid system calls
    mocker.patch("artifex.core.decorators.logging.psutil.virtual_memory", return_value=mocker.MagicMock(percent=50.0))
    mock_process = mocker.MagicMock()
    mock_process.cpu_percent.return_value = 25.0
    mocker.patch("artifex.core.decorators.logging.psutil.Process", return_value=mock_process)
    mocker.patch("artifex.core.decorators.logging.psutil.cpu_count", return_value=4)
    mocker.patch("artifex.core.decorators.logging.time.time", side_effect=[100.0, 101.0])
    
    # Mock pipeline to return expected output
    mock_classifier = mock_pipeline.return_value
    mock_classifier.return_value = [{"label": "positive", "score": 0.95}]
    
    # Call the method
    result = concrete_model("test input text")
    
    # Verify the log file was created
    assert log_file.exists()
    
    # Read and verify log entry
    log_content = log_file.read_text().strip()
    log_entry = json.loads(log_content)
    
    # Verify log entry contains expected fields
    assert log_entry["entry_type"] == "inference"
    # Model name will be the concrete implementation class name
    assert "model" in log_entry
    assert log_entry["model"] in ["ClassificationModel", "ConcreteClassificationModel"]
    assert "inputs" in log_entry
    assert "output" in log_entry
    assert "inference_duration_seconds" in log_entry
    assert "cpu_usage_percent" in log_entry
    assert "ram_usage_percent" in log_entry
    assert "input_token_count" in log_entry
    assert "timestamp" in log_entry
    
    # Verify result is correct - returns list of ClassificationResponse objects
    assert isinstance(result, list)
    assert len(result) == 1
    assert result[0].label == "positive"
    assert result[0].score == 0.95


@pytest.mark.unit
def test_call_with_disable_logging_prevents_logging(
    concrete_model: ClassificationModel,
    mock_pipeline: MockerFixture,
    mocker: MockerFixture,
    tmp_path
):
    """
    Test that __call__ does not log when disable_logging=True is passed.
    
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        mock_pipeline (MockerFixture): Mocked pipeline function.
        mocker (MockerFixture): The pytest-mock fixture for mocking.
        tmp_path: Pytest fixture for temporary directory.
    """
    import json
    from pathlib import Path
    
    log_file = tmp_path / "inference.log"
    
    # Mock the config paths
    mocker.patch("artifex.core.decorators.logging.config.INFERENCE_LOGS_PATH", str(log_file))
    
    # Mock pipeline to return expected output
    mock_classifier = mock_pipeline.return_value
    mock_classifier.return_value = [{"label": "negative", "score": 0.85}]
    
    # Call the method with disable_logging=True
    result = concrete_model("test input text", disable_logging=True)
    
    # Verify the log file was NOT created
    assert not log_file.exists()
    
    # Verify result is still correct
    assert isinstance(result, list)
    assert len(result) == 1
    assert result[0].label == "negative"
    assert result[0].score == 0.85


import pytest
import json
from pytest_mock import MockerFixture

from artifex.core.decorators.logging import track_training_calls


@pytest.fixture
def setup_mocks(mocker, tmp_path):
    """Common mocking setup for track_training_calls warning tests."""
    log_file = tmp_path / "training.log"
    warnings_file = tmp_path / "warnings.log"
    
    mocker.patch("artifex.core.decorators.logging.config.TRAINING_LOGS_PATH", str(log_file))
    mocker.patch("artifex.core.decorators.logging.config.WARNINGS_LOGS_PATH", str(warnings_file))
    mocker.patch("artifex.core.decorators.logging._calculate_daily_training_aggregates")
    mocker.patch("artifex.core.decorators.logging._to_json", side_effect=lambda x: x)
    mocker.patch("artifex.core.decorators.logging.psutil.virtual_memory", return_value=mocker.MagicMock(percent=50.0))
    
    mock_process = mocker.MagicMock()
    mock_process.cpu_percent.return_value = 25.0
    mocker.patch("artifex.core.decorators.logging.psutil.Process", return_value=mock_process)
    
    mocker.patch("artifex.core.decorators.logging.psutil.cpu_count", return_value=4)
    
    return log_file, warnings_file


@pytest.mark.unit
def test_track_training_calls_logs_high_loss_warning(mocker, setup_mocks):
    """
    Test that track_training_calls logs warning when train_loss > 1.0.
    """
    log_file, warnings_file = setup_mocks
    
    mocker.patch("artifex.core.decorators.logging.time.time", side_effect=[100.0, 150.0])
    mocker.patch("artifex.core.decorators.logging._extract_train_metrics", return_value={"train_loss": 1.5, "epoch": 3})
    
    @track_training_calls
    def test_func(self, x):
        return {"metrics": {"train_loss": 1.5}}
    
    class TestClass:
        pass
    
    instance = TestClass()
    test_func(instance, 5)
    
    assert warnings_file.exists()
    warning_content = warnings_file.read_text()
    warning_entry = json.loads(warning_content.strip())
    
    assert warning_entry["entry_type"] == "high_training_loss_warning"
    assert "1.5" in warning_entry["warning_message"]
    assert "exceeded 1.0 threshold" in warning_entry["warning_message"]


@pytest.mark.unit
def test_track_training_calls_no_warning_for_low_loss(mocker, setup_mocks):
    """
    Test that track_training_calls does NOT log warning when train_loss <= 1.0.
    """
    log_file, warnings_file = setup_mocks
    
    mocker.patch("artifex.core.decorators.logging.time.time", side_effect=[100.0, 150.0])
    mocker.patch("artifex.core.decorators.logging._extract_train_metrics", return_value={"train_loss": 0.3, "epoch": 3})
    
    @track_training_calls
    def test_func(self, x):
        return {"metrics": {"train_loss": 0.3}}
    
    class TestClass:
        pass
    
    instance = TestClass()
    test_func(instance, 5)
    
    assert not warnings_file.exists()


@pytest.mark.unit
def test_track_training_calls_logs_high_loss_warning_with_loss_metric(mocker, setup_mocks):
    """
    Test that track_training_calls detects high loss with 'loss' key.
    """
    log_file, warnings_file = setup_mocks
    
    mocker.patch("artifex.core.decorators.logging.time.time", side_effect=[100.0, 150.0])
    mocker.patch("artifex.core.decorators.logging._extract_train_metrics", return_value={"loss": 2.0, "epoch": 3})
    
    @track_training_calls
    def test_func(self, x):
        return {"metrics": {"loss": 2.0}}
    
    class TestClass:
        pass
    
    instance = TestClass()
    test_func(instance, 5)
    
    assert warnings_file.exists()
    warning_content = warnings_file.read_text()
    warning_entry = json.loads(warning_content.strip())
    
    assert warning_entry["entry_type"] == "high_training_loss_warning"
    assert "2.0" in warning_entry["warning_message"]


@pytest.mark.unit
def test_track_training_calls_logs_high_loss_warning_with_eval_loss(mocker, setup_mocks):
    """
    Test that track_training_calls detects high loss with 'eval_loss' key.
    """
    log_file, warnings_file = setup_mocks
    
    mocker.patch("artifex.core.decorators.logging.time.time", side_effect=[100.0, 150.0])
    mocker.patch("artifex.core.decorators.logging._extract_train_metrics", return_value={"eval_loss": 1.8, "epoch": 3})
    
    @track_training_calls
    def test_func(self, x):
        return {"metrics": {"eval_loss": 1.8}}
    
    class TestClass:
        pass
    
    instance = TestClass()
    test_func(instance, 5)
    
    assert warnings_file.exists()
    warning_content = warnings_file.read_text()
    warning_entry = json.loads(warning_content.strip())
    
    assert warning_entry["entry_type"] == "high_training_loss_warning"
    assert "1.8" in warning_entry["warning_message"]


@pytest.mark.unit
def test_track_training_calls_logs_slow_training_warning(mocker, setup_mocks):
    """
    Test that track_training_calls logs warning when training duration > 300 seconds.
    """
    log_file, warnings_file = setup_mocks
    
    mocker.patch("artifex.core.decorators.logging.time.time", side_effect=[100.0, 450.0])  # 350 seconds
    mocker.patch("artifex.core.decorators.logging._extract_train_metrics", return_value={"train_loss": 0.3, "epoch": 3})
    
    @track_training_calls
    def test_func(self, x):
        return {"metrics": {"train_loss": 0.3}}
    
    class TestClass:
        pass
    
    instance = TestClass()
    test_func(instance, 5)
    
    assert warnings_file.exists()
    warning_content = warnings_file.read_text()
    warning_entry = json.loads(warning_content.strip())
    
    assert warning_entry["entry_type"] == "slow_training_warning"
    assert "350" in warning_entry["warning_message"]
    assert "exceeded 300 second threshold" in warning_entry["warning_message"]


@pytest.mark.unit
def test_track_training_calls_no_warning_for_fast_training(mocker, setup_mocks):
    """
    Test that track_training_calls does NOT log warning when training duration <= 300 seconds.
    """
    log_file, warnings_file = setup_mocks
    
    mocker.patch("artifex.core.decorators.logging.time.time", side_effect=[100.0, 250.0])  # 150 seconds
    mocker.patch("artifex.core.decorators.logging._extract_train_metrics", return_value={"train_loss": 0.3, "epoch": 3})
    
    @track_training_calls
    def test_func(self, x):
        return {"metrics": {"train_loss": 0.3}}
    
    class TestClass:
        pass
    
    instance = TestClass()
    test_func(instance, 5)
    
    assert not warnings_file.exists()


@pytest.mark.unit
def test_track_training_calls_logs_low_throughput_warning(mocker, setup_mocks):
    """
    Test that track_training_calls logs warning when train_samples_per_second < 1.0.
    """
    log_file, warnings_file = setup_mocks
    
    mocker.patch("artifex.core.decorators.logging.time.time", side_effect=[100.0, 150.0])
    mocker.patch("artifex.core.decorators.logging._extract_train_metrics", return_value={
        "train_loss": 0.3,
        "train_samples_per_second": 0.5,
        "epoch": 3
    })
    
    @track_training_calls
    def test_func(self, x):
        return {"metrics": {"train_samples_per_second": 0.5}}
    
    class TestClass:
        pass
    
    instance = TestClass()
    test_func(instance, 5)
    
    assert warnings_file.exists()
    warning_content = warnings_file.read_text()
    warning_entry = json.loads(warning_content.strip())
    
    assert warning_entry["entry_type"] == "low_training_throughput_warning"
    assert "0.5" in warning_entry["warning_message"]
    assert "below 1.0 threshold" in warning_entry["warning_message"]


@pytest.mark.unit
def test_track_training_calls_no_warning_for_good_throughput(mocker, setup_mocks):
    """
    Test that track_training_calls does NOT log warning when train_samples_per_second >= 1.0.
    """
    log_file, warnings_file = setup_mocks
    
    mocker.patch("artifex.core.decorators.logging.time.time", side_effect=[100.0, 150.0])
    mocker.patch("artifex.core.decorators.logging._extract_train_metrics", return_value={
        "train_loss": 0.3,
        "train_samples_per_second": 5.0,
        "epoch": 3
    })
    
    @track_training_calls
    def test_func(self, x):
        return {"metrics": {"train_samples_per_second": 5.0}}
    
    class TestClass:
        pass
    
    instance = TestClass()
    test_func(instance, 5)
    
    assert not warnings_file.exists()


@pytest.mark.unit
def test_track_training_calls_multiple_warnings_logged(mocker, setup_mocks):
    """
    Test that track_training_calls logs multiple warnings when multiple conditions are met.
    """
    log_file, warnings_file = setup_mocks
    
    mocker.patch("artifex.core.decorators.logging.time.time", side_effect=[100.0, 500.0])  # 400 seconds (slow)
    mocker.patch("artifex.core.decorators.logging._extract_train_metrics", return_value={
        "train_loss": 1.8,  # High loss
        "train_samples_per_second": 0.3,  # Low throughput
        "epoch": 3
    })
    
    @track_training_calls
    def test_func(self, x):
        return {"metrics": {"train_loss": 1.8, "train_samples_per_second": 0.3}}
    
    class TestClass:
        pass
    
    instance = TestClass()
    test_func(instance, 5)
    
    assert warnings_file.exists()
    warning_lines = warnings_file.read_text().strip().split("\n")
    assert len(warning_lines) == 3  # 3 warnings
    
    warning_types = [json.loads(line)["entry_type"] for line in warning_lines]
    assert "high_training_loss_warning" in warning_types
    assert "slow_training_warning" in warning_types
    assert "low_training_throughput_warning" in warning_types


@pytest.mark.unit
def test_track_training_calls_warning_includes_all_training_data(mocker, setup_mocks):
    """
    Test that warning entry includes all the same data as regular training entry.
    """
    log_file, warnings_file = setup_mocks
    
    mocker.patch("artifex.core.decorators.logging.time.time", side_effect=[100.0, 450.0])  # Slow training
    mocker.patch("artifex.core.decorators.logging._extract_train_metrics", return_value={
        "train_loss": 0.3,
        "epoch": 3
    })
    
    @track_training_calls
    def test_func(self, x):
        return {"metrics": {"train_loss": 0.3}}
    
    class TestClass:
        pass
    
    instance = TestClass()
    test_func(instance, 5)
    
    # Read both log files
    training_entry = json.loads(log_file.read_text().strip())
    warning_entry = json.loads(warnings_file.read_text().strip())
    
    # Warning entry should have all the same fields as training entry
    assert warning_entry["timestamp"] == training_entry["timestamp"]
    assert warning_entry["model"] == training_entry["model"]
    assert warning_entry["training_duration_seconds"] == training_entry["training_duration_seconds"]
    assert warning_entry["cpu_usage_percent"] == training_entry["cpu_usage_percent"]
    assert warning_entry["ram_usage_percent"] == training_entry["ram_usage_percent"]
    assert warning_entry["train_results"] == training_entry["train_results"]
    
    # Plus the warning-specific fields
    assert warning_entry["entry_type"] == "slow_training_warning"
    assert "exceeded 300 second threshold" in warning_entry["warning_message"]


@pytest.mark.unit
def test_track_training_calls_no_warnings_for_good_training(mocker, setup_mocks):
    """
    Test that track_training_calls does NOT log warnings when all metrics are good.
    """
    log_file, warnings_file = setup_mocks
    
    mocker.patch("artifex.core.decorators.logging.time.time", side_effect=[100.0, 250.0])  # Fast: 150 seconds
    mocker.patch("artifex.core.decorators.logging._extract_train_metrics", return_value={
        "train_loss": 0.25,  # Low loss
        "train_samples_per_second": 8.5,  # Good throughput
        "epoch": 3
    })
    
    @track_training_calls
    def test_func(self, x):
        return {"metrics": {"train_loss": 0.25, "train_samples_per_second": 8.5}}
    
    class TestClass:
        pass
    
    instance = TestClass()
    test_func(instance, 5)
    
    # Should have training log but no warnings
    assert log_file.exists()
    assert not warnings_file.exists()


@pytest.mark.unit
def test_track_training_calls_no_warning_when_no_train_results(mocker, setup_mocks):
    """
    Test that track_training_calls does NOT log warnings when train_results is None.
    """
    log_file, warnings_file = setup_mocks
    
    mocker.patch("artifex.core.decorators.logging.time.time", side_effect=[100.0, 150.0])
    mocker.patch("artifex.core.decorators.logging._extract_train_metrics", return_value=None)
    
    @track_training_calls
    def test_func(self, x):
        return None
    
    class TestClass:
        pass
    
    instance = TestClass()
    test_func(instance, 5)
    
    # No warnings should be logged
    assert not warnings_file.exists()



import pytest
import json
from pytest_mock import MockerFixture

from artifex.core.decorators.logging import track_training_calls


@pytest.fixture
def setup_training_mocks_with_shipping(mocker, tmp_path):
    """Common mocking setup for training warning shipping tests."""
    log_file = tmp_path / "training.log"
    warnings_file = tmp_path / "warnings.log"
    
    mocker.patch("artifex.core.decorators.logging.config.TRAINING_LOGS_PATH", str(log_file))
    mocker.patch("artifex.core.decorators.logging.config.WARNINGS_LOGS_PATH", str(warnings_file))
    mocker.patch("artifex.core.decorators.logging._calculate_daily_training_aggregates")
    mocker.patch("artifex.core.decorators.logging._to_json", side_effect=lambda x: x)
    mocker.patch("artifex.core.decorators.logging._extract_train_metrics", side_effect=lambda x: x)
    mocker.patch("artifex.core.decorators.logging.psutil.virtual_memory", return_value=mocker.MagicMock(percent=60.0))
    
    mock_process = mocker.MagicMock()
    mock_process.cpu_percent.return_value = 30.0
    mocker.patch("artifex.core.decorators.logging.psutil.Process", return_value=mock_process)
    
    mocker.patch("artifex.core.decorators.logging.psutil.cpu_count", return_value=4)
    mocker.patch("artifex.core.decorators.logging.time.time", side_effect=[100.0, 150.0])
    
    # Mock ship_log function
    mock_ship_log = mocker.patch("artifex.core.decorators.logging.ship_log")
    
    return {
        "log_file": log_file,
        "warnings_file": warnings_file,
        "mock_ship_log": mock_ship_log
    }


@pytest.mark.unit
def test_training_warning_shipping_high_loss(setup_training_mocks_with_shipping):
    """
    Test that high training loss warning is shipped to cloud.
    """
    mocks = setup_training_mocks_with_shipping
    
    @track_training_calls
    def train_func(self, data):
        return {"train_loss": 2.5}  # High loss > 1.0
    
    class TestModel:
        pass
    
    instance = TestModel()
    train_func(instance, "training_data")
    
    # Find the warning call
    warning_calls = [call for call in mocks["mock_ship_log"].call_args_list 
                     if call[0][1] == "training-warnings"]
    
    assert len(warning_calls) == 1
    warning_data = warning_calls[0][0][0]
    assert warning_data["warning_type"] == "high_training_loss_warning"
    assert "2.5" in warning_data["warning_message"]


@pytest.mark.unit
def test_training_warning_shipping_high_loss_different_metric_names(mocker, setup_training_mocks_with_shipping):
    """
    Test that high training loss warning works with different metric names.
    """
    mocks = setup_training_mocks_with_shipping
    
    # Test with "loss" key
    @track_training_calls
    def train_func1(self, data):
        return {"loss": 1.8}
    
    class TestModel:
        pass
    
    instance = TestModel()
    train_func1(instance, "data")
    
    warning_calls = [call for call in mocks["mock_ship_log"].call_args_list 
                     if call[0][1] == "training-warnings"]
    assert len(warning_calls) == 1
    assert warning_calls[0][0][0]["warning_type"] == "high_training_loss_warning"
    
    # Reset mock
    mocks["mock_ship_log"].reset_mock()
    mocker.patch("artifex.core.decorators.logging.time.time", side_effect=[100.0, 150.0])
    
    # Test with "eval_loss" key
    @track_training_calls
    def train_func2(self, data):
        return {"eval_loss": 1.5}
    
    train_func2(instance, "data")
    
    warning_calls = [call for call in mocks["mock_ship_log"].call_args_list 
                     if call[0][1] == "training-warnings"]
    assert len(warning_calls) == 1
    assert warning_calls[0][0][0]["warning_type"] == "high_training_loss_warning"


@pytest.mark.unit
def test_training_warning_shipping_slow_training(mocker, tmp_path):
    """
    Test that slow training warning is shipped to cloud.
    """
    log_file = tmp_path / "training.log"
    warnings_file = tmp_path / "warnings.log"
    
    mocker.patch("artifex.core.decorators.logging.config.TRAINING_LOGS_PATH", str(log_file))
    mocker.patch("artifex.core.decorators.logging.config.WARNINGS_LOGS_PATH", str(warnings_file))
    mocker.patch("artifex.core.decorators.logging._calculate_daily_training_aggregates")
    mocker.patch("artifex.core.decorators.logging._to_json", side_effect=lambda x: x)
    mocker.patch("artifex.core.decorators.logging._extract_train_metrics", side_effect=lambda x: x)
    mocker.patch("artifex.core.decorators.logging.psutil.virtual_memory", return_value=mocker.MagicMock(percent=60.0))
    
    mock_process = mocker.MagicMock()
    mock_process.cpu_percent.return_value = 30.0
    mocker.patch("artifex.core.decorators.logging.psutil.Process", return_value=mock_process)
    
    mocker.patch("artifex.core.decorators.logging.psutil.cpu_count", return_value=4)
    # Duration of 350 seconds (> 300 second threshold)
    mocker.patch("artifex.core.decorators.logging.time.time", side_effect=[100.0, 450.0])
    
    mock_ship_log = mocker.patch("artifex.core.decorators.logging.ship_log")
    
    @track_training_calls
    def train_func(self, data):
        return {"train_loss": 0.5}
    
    class TestModel:
        pass
    
    instance = TestModel()
    train_func(instance, "data")
    
    # Find the warning call
    warning_calls = [call for call in mock_ship_log.call_args_list 
                     if call[0][1] == "training-warnings"]
    
    assert len(warning_calls) == 1
    warning_data = warning_calls[0][0][0]
    assert warning_data["warning_type"] == "slow_training_warning"
    assert "350" in warning_data["warning_message"]


@pytest.mark.unit
def test_training_warning_shipping_low_throughput(setup_training_mocks_with_shipping):
    """
    Test that low training throughput warning is shipped to cloud.
    """
    mocks = setup_training_mocks_with_shipping
    
    @track_training_calls
    def train_func(self, data):
        return {"train_samples_per_second": 0.5}  # < 1.0 threshold
    
    class TestModel:
        pass
    
    instance = TestModel()
    train_func(instance, "data")
    
    # Find the warning call
    warning_calls = [call for call in mocks["mock_ship_log"].call_args_list 
                     if call[0][1] == "training-warnings"]
    
    assert len(warning_calls) == 1
    warning_data = warning_calls[0][0][0]
    assert warning_data["warning_type"] == "low_training_throughput_warning"
    assert "0.5" in warning_data["warning_message"]


@pytest.mark.unit
def test_training_warning_shipping_multiple_warnings(mocker, tmp_path):
    """
    Test that multiple training warnings can be shipped in a single call.
    """
    log_file = tmp_path / "training.log"
    warnings_file = tmp_path / "warnings.log"
    
    mocker.patch("artifex.core.decorators.logging.config.TRAINING_LOGS_PATH", str(log_file))
    mocker.patch("artifex.core.decorators.logging.config.WARNINGS_LOGS_PATH", str(warnings_file))
    mocker.patch("artifex.core.decorators.logging._calculate_daily_training_aggregates")
    mocker.patch("artifex.core.decorators.logging._to_json", side_effect=lambda x: x)
    mocker.patch("artifex.core.decorators.logging._extract_train_metrics", side_effect=lambda x: x)
    mocker.patch("artifex.core.decorators.logging.psutil.virtual_memory", return_value=mocker.MagicMock(percent=60.0))
    
    mock_process = mocker.MagicMock()
    mock_process.cpu_percent.return_value = 30.0
    mocker.patch("artifex.core.decorators.logging.psutil.Process", return_value=mock_process)
    
    mocker.patch("artifex.core.decorators.logging.psutil.cpu_count", return_value=4)
    # Slow training (350 seconds)
    mocker.patch("artifex.core.decorators.logging.time.time", side_effect=[100.0, 450.0])
    
    mock_ship_log = mocker.patch("artifex.core.decorators.logging.ship_log")
    
    @track_training_calls
    def train_func(self, data):
        # High loss + slow training + low throughput
        return {
            "train_loss": 1.5,
            "train_samples_per_second": 0.3
        }
    
    class TestModel:
        pass
    
    instance = TestModel()
    train_func(instance, "data")
    
    # Find all warning calls
    warning_calls = [call for call in mock_ship_log.call_args_list 
                     if call[0][1] == "training-warnings"]
    
    # Should have 3 warnings
    assert len(warning_calls) == 3
    
    warning_types = [call[0][0]["warning_type"] for call in warning_calls]
    assert "high_training_loss_warning" in warning_types
    assert "slow_training_warning" in warning_types
    assert "low_training_throughput_warning" in warning_types


@pytest.mark.unit
def test_training_warning_shipping_inputs_args_serialized_as_json_string(setup_training_mocks_with_shipping):
    """
    Test that inputs.args is serialized as a JSON string for API compatibility.
    """
    mocks = setup_training_mocks_with_shipping
    
    @track_training_calls
    def train_func(self, dataset, epochs):
        return {"train_loss": 1.5}
    
    class TestModel:
        pass
    
    instance = TestModel()
    train_func(instance, "my_dataset", 10)
    
    # Find the warning call
    warning_calls = [call for call in mocks["mock_ship_log"].call_args_list 
                     if call[0][1] == "training-warnings"]
    
    assert len(warning_calls) == 1
    warning_data = warning_calls[0][0][0]
    
    # Verify inputs.args is a JSON string
    assert "inputs" in warning_data
    assert "args" in warning_data["inputs"]
    args_value = warning_data["inputs"]["args"]
    assert isinstance(args_value, str)
    # Should be valid JSON
    parsed_args = json.loads(args_value)
    assert parsed_args == ["my_dataset", 10]


@pytest.mark.unit
def test_training_warning_shipping_contains_all_metadata(setup_training_mocks_with_shipping):
    """
    Test that shipped training warnings contain all required metadata.
    """
    mocks = setup_training_mocks_with_shipping
    
    @track_training_calls
    def train_func(self, data):
        return {"train_loss": 2.0}
    
    class TestModel:
        pass
    
    instance = TestModel()
    train_func(instance, "data")
    
    # Find the warning call
    warning_calls = [call for call in mocks["mock_ship_log"].call_args_list 
                     if call[0][1] == "training-warnings"]
    
    warning_data = warning_calls[0][0][0]
    
    # Verify all required fields are present
    assert "warning_type" in warning_data
    assert "warning_message" in warning_data
    assert "timestamp" in warning_data
    assert "model" in warning_data
    assert "training_duration_seconds" in warning_data
    assert "cpu_usage_percent" in warning_data
    assert "ram_usage_percent" in warning_data
    assert "inputs" in warning_data
    assert "train_results" in warning_data


@pytest.mark.unit
def test_training_warning_shipping_no_warning_for_good_training(setup_training_mocks_with_shipping):
    """
    Test that no warnings are shipped for a good training run.
    """
    mocks = setup_training_mocks_with_shipping
    
    @track_training_calls
    def train_func(self, data):
        return {
            "train_loss": 0.3,
            "train_samples_per_second": 10.0
        }
    
    class TestModel:
        pass
    
    instance = TestModel()
    train_func(instance, "data")
    
    # Find warning calls
    warning_calls = [call for call in mocks["mock_ship_log"].call_args_list 
                     if call[0][1] == "training-warnings"]
    
    # Should have no warnings
    assert len(warning_calls) == 0


@pytest.mark.unit
def test_training_warning_shipping_creates_warnings_log_file(setup_training_mocks_with_shipping):
    """
    Test that warnings log file is created when training warnings are generated.
    """
    mocks = setup_training_mocks_with_shipping
    
    @track_training_calls
    def train_func(self, data):
        return {"train_loss": 1.5}
    
    class TestModel:
        pass
    
    instance = TestModel()
    train_func(instance, "data")
    
    # Verify warnings file was created
    assert mocks["warnings_file"].exists()
    
    # Verify content
    content = mocks["warnings_file"].read_text()
    warning_entry = json.loads(content.strip())
    assert warning_entry["warning_type"] == "high_training_loss_warning"


@pytest.mark.unit
def test_training_warning_shipping_correct_log_type(setup_training_mocks_with_shipping):
    """
    Test that training warnings are shipped with correct log_type parameter.
    """
    mocks = setup_training_mocks_with_shipping
    
    @track_training_calls
    def train_func(self, data):
        return {"train_loss": 1.5}
    
    class TestModel:
        pass
    
    instance = TestModel()
    train_func(instance, "data")
    
    # Find the warning call
    warning_calls = [call for call in mocks["mock_ship_log"].call_args_list 
                     if call[0][1] == "training-warnings"]
    
    assert len(warning_calls) == 1
    # Verify the log_type parameter
    assert warning_calls[0][0][1] == "training-warnings"


@pytest.mark.unit
def test_training_warning_shipping_threshold_exactly_1_0(setup_training_mocks_with_shipping):
    """
    Test that training loss of exactly 1.0 does NOT trigger a warning.
    """
    mocks = setup_training_mocks_with_shipping
    
    @track_training_calls
    def train_func(self, data):
        return {"train_loss": 1.0}  # Exactly at threshold
    
    class TestModel:
        pass
    
    instance = TestModel()
    train_func(instance, "data")
    
    # Find warning calls
    warning_calls = [call for call in mocks["mock_ship_log"].call_args_list 
                     if call[0][1] == "training-warnings"]
    
    # Should have no warnings (> 1.0 triggers warning, not >= 1.0)
    assert len(warning_calls) == 0


@pytest.mark.unit
def test_training_warning_shipping_threshold_exactly_300_seconds(mocker, tmp_path):
    """
    Test that training duration of exactly 300 seconds does NOT trigger a warning.
    """
    log_file = tmp_path / "training.log"
    warnings_file = tmp_path / "warnings.log"
    
    mocker.patch("artifex.core.decorators.logging.config.TRAINING_LOGS_PATH", str(log_file))
    mocker.patch("artifex.core.decorators.logging.config.WARNINGS_LOGS_PATH", str(warnings_file))
    mocker.patch("artifex.core.decorators.logging._calculate_daily_training_aggregates")
    mocker.patch("artifex.core.decorators.logging._to_json", side_effect=lambda x: x)
    mocker.patch("artifex.core.decorators.logging._extract_train_metrics", side_effect=lambda x: x)
    mocker.patch("artifex.core.decorators.logging.psutil.virtual_memory", return_value=mocker.MagicMock(percent=60.0))
    
    mock_process = mocker.MagicMock()
    mock_process.cpu_percent.return_value = 30.0
    mocker.patch("artifex.core.decorators.logging.psutil.Process", return_value=mock_process)
    
    mocker.patch("artifex.core.decorators.logging.psutil.cpu_count", return_value=4)
    # Exactly 300 seconds
    mocker.patch("artifex.core.decorators.logging.time.time", side_effect=[100.0, 400.0])
    
    mock_ship_log = mocker.patch("artifex.core.decorators.logging.ship_log")
    
    @track_training_calls
    def train_func(self, data):
        return {"train_loss": 0.5}
    
    class TestModel:
        pass
    
    instance = TestModel()
    train_func(instance, "data")
    
    # Find warning calls
    warning_calls = [call for call in mock_ship_log.call_args_list 
                     if call[0][1] == "training-warnings"]
    
    # Should have no warnings (> 300 triggers warning, not >= 300)
    assert len(warning_calls) == 0


@pytest.mark.unit
def test_training_warning_shipping_threshold_exactly_1_samples_per_second(setup_training_mocks_with_shipping):
    """
    Test that throughput of exactly 1.0 samples/second does NOT trigger a warning.
    """
    mocks = setup_training_mocks_with_shipping
    
    @track_training_calls
    def train_func(self, data):
        return {"train_samples_per_second": 1.0}  # Exactly at threshold
    
    class TestModel:
        pass
    
    instance = TestModel()
    train_func(instance, "data")
    
    # Find warning calls
    warning_calls = [call for call in mocks["mock_ship_log"].call_args_list 
                     if call[0][1] == "training-warnings"]
    
    # Should have no warnings (< 1.0 triggers warning, not <= 1.0)
    assert len(warning_calls) == 0


@pytest.mark.unit
def test_training_warning_shipping_with_kwargs(setup_training_mocks_with_shipping):
    """
    Test that warnings are shipped correctly when train function is called with kwargs.
    """
    mocks = setup_training_mocks_with_shipping
    
    @track_training_calls
    def train_func(self, dataset, epochs=10, learning_rate=0.001):
        return {"train_loss": 1.5}
    
    class TestModel:
        pass
    
    instance = TestModel()
    train_func(instance, "my_data", epochs=20, learning_rate=0.0001)
    
    # Find the warning call
    warning_calls = [call for call in mocks["mock_ship_log"].call_args_list 
                     if call[0][1] == "training-warnings"]
    
    assert len(warning_calls) == 1
    warning_data = warning_calls[0][0][0]
    
    # Verify kwargs are in inputs
    assert "inputs" in warning_data
    assert "kwargs" in warning_data["inputs"]
    assert warning_data["inputs"]["kwargs"]["epochs"] == 20
    assert warning_data["inputs"]["kwargs"]["learning_rate"] == 0.0001


@pytest.mark.unit
def test_training_warning_shipping_writes_to_file_before_shipping(setup_training_mocks_with_shipping):
    """
    Test that warnings are written to file before being shipped to cloud.
    """
    mocks = setup_training_mocks_with_shipping
    
    @track_training_calls
    def train_func(self, data):
        return {"train_loss": 2.0}
    
    class TestModel:
        pass
    
    instance = TestModel()
    train_func(instance, "data")
    
    # Verify file exists
    assert mocks["warnings_file"].exists()
    
    # Verify shipping occurred
    warning_calls = [call for call in mocks["mock_ship_log"].call_args_list 
                     if call[0][1] == "training-warnings"]
    assert len(warning_calls) == 1


@pytest.mark.unit
def test_training_warning_shipping_with_disable_logging(mocker, tmp_path):
    """
    Test that no warnings are shipped when disable_logging=True.
    """
    mock_ship_log = mocker.patch("artifex.core.decorators.logging.ship_log")
    
    @track_training_calls
    def train_func(self, data):
        return {"train_loss": 5.0}  # Would trigger warning
    
    class TestModel:
        pass
    
    instance = TestModel()
    train_func(instance, "data", disable_logging=True)
    
    # Should not call ship_log at all
    mock_ship_log.assert_not_called()



from synthex import Synthex
import pytest
from pytest_mock import MockerFixture
import torch

from artifex.models import Reranker
from artifex.config import config


@pytest.fixture(scope="function", autouse=True)
def mock_dependencies(mocker: MockerFixture):
    """
    Fixture to mock all external dependencies before any test runs.
    This fixture runs automatically for all tests in this module.
    Args:
        mocker (MockerFixture): The pytest-mock fixture for mocking.
    """
    
    # Mock config
    mocker.patch.object(config, "RERANKER_HF_BASE_MODEL", "mock-reranker-model")
    mocker.patch.object(config, "RERANKER_TOKENIZER_MAX_LENGTH", 512)
    
    # Mock AutoTokenizer at the module where it's used
    mock_tokenizer = mocker.MagicMock()
    mocker.patch(
        "artifex.models.reranker.reranker.AutoTokenizer.from_pretrained",
        return_value=mock_tokenizer
    )
    
    # Mock AutoModelForSequenceClassification at the module where it's used
    mock_model = mocker.MagicMock()
    mocker.patch(
        "artifex.models.reranker.reranker.AutoModelForSequenceClassification.from_pretrained",
        return_value=mock_model
    )


@pytest.fixture
def mock_synthex(mocker: MockerFixture) -> Synthex:
    """
    Fixture to create a mock Synthex instance.
    Args:
        mocker (MockerFixture): The pytest-mock fixture for mocking.
    Returns:
        Synthex: A mocked Synthex instance.
    """
    
    return mocker.MagicMock(spec=Synthex)


@pytest.fixture
def mock_reranker(mocker: MockerFixture, mock_synthex: Synthex) -> Reranker:
    """
    Fixture to create a Reranker instance with mocked dependencies.
    Args:
        mocker (MockerFixture): The pytest-mock fixture for mocking.
        mock_synthex (Synthex): A mocked Synthex instance.
    Returns:
        Reranker: An instance of the Reranker model with mocked dependencies.
    """
    
    reranker = Reranker(mock_synthex)
    
    # Mock the tokenizer to return proper inputs
    mock_tokenizer_output = {
        "input_ids": torch.tensor([[1, 2, 3]]),
        "attention_mask": torch.tensor([[1, 1, 1]])
    }
    reranker._tokenizer.return_value = mock_tokenizer_output
    
    # Mock the model output
    mock_model_output = mocker.MagicMock()
    reranker._model.return_value = mock_model_output
    
    return reranker


@pytest.mark.unit
def test_call_with_single_document(
    mock_reranker: Reranker, mocker: MockerFixture
):
    """
    Test that __call__ works correctly with a single document string.
    Args:
        mock_reranker (Reranker): The Reranker instance with mocked dependencies.
        mocker (MockerFixture): The pytest-mock fixture for mocking.
    """
    
    query = "What is machine learning?"
    document = "Machine learning is a subset of artificial intelligence."
    
    # Mock model output
    mock_logits = torch.tensor([[0.85]])
    mock_output = mocker.MagicMock()
    mock_output.logits = mock_logits
    mock_reranker._model.return_value = mock_output
    
    result = mock_reranker(query, document)
    
    # Should return a list with one tuple
    assert isinstance(result, list)
    assert len(result) == 1
    assert isinstance(result[0], tuple)
    assert result[0][0] == document
    assert isinstance(result[0][1], float)


@pytest.mark.unit
def test_call_with_multiple_documents(
    mock_reranker: Reranker, mocker: MockerFixture
):
    """
    Test that __call__ works correctly with multiple documents.
    Args:
        mock_reranker (Reranker): The Reranker instance with mocked dependencies.
        mocker (MockerFixture): The pytest-mock fixture for mocking.
    """
    
    query = "What is Python?"
    documents = [
        "Python is a programming language.",
        "The python is a snake.",
        "Python was created by Guido van Rossum."
    ]
    
    # Mock model output with different scores
    mock_logits = torch.tensor([[0.9], [0.2], [0.75]])
    mock_output = mocker.MagicMock()
    mock_output.logits = mock_logits
    mock_reranker._model.return_value = mock_output
    
    result = mock_reranker(query, documents)
    
    # Should return a list with three tuples
    assert isinstance(result, list)
    assert len(result) == 3
    
    # Check that all documents are present
    result_docs = [doc for doc, _ in result]
    assert set(result_docs) == set(documents)
    
    # Check that results are sorted by score (descending)
    scores = [score for _, score in result]
    assert scores == sorted(scores, reverse=True)


@pytest.mark.unit
def test_call_documents_sorted_by_score(
    mock_reranker: Reranker, mocker: MockerFixture
):
    """
    Test that documents are correctly sorted by relevance score in descending order.
    Args:
        mock_reranker (Reranker): The Reranker instance with mocked dependencies.
        mocker (MockerFixture): The pytest-mock fixture for mocking.
    """
    
    query = "climate change"
    documents = ["Doc A", "Doc B", "Doc C"]
    
    # Mock model output with specific scores
    mock_logits = torch.tensor([[0.3], [0.9], [0.6]])
    mock_output = mocker.MagicMock()
    mock_output.logits = mock_logits
    mock_reranker._model.return_value = mock_output
    
    result = mock_reranker(query, documents)
    
    # Check ordering: Doc B (0.9), Doc C (0.6), Doc A (0.3)
    assert result[0][0] == "Doc B"
    assert result[1][0] == "Doc C"
    assert result[2][0] == "Doc A"
    
    assert result[0][1] > result[1][1] > result[2][1]


@pytest.mark.unit
def test_call_tokenizer_called_correctly(
    mock_reranker: Reranker, mocker: MockerFixture
):
    """
    Test that the tokenizer is called with correct parameters.
    Args:
        mock_reranker (Reranker): The Reranker instance with mocked dependencies.
        mocker (MockerFixture): The pytest-mock fixture for mocking.
    """
    
    query = "test query"
    documents = ["doc1", "doc2"]
    
    # Mock model output
    mock_logits = torch.tensor([[0.5], [0.7]])
    mock_output = mocker.MagicMock()
    mock_output.logits = mock_logits
    mock_reranker._model.return_value = mock_output
    
    mock_reranker(query, documents)
    
    # Verify tokenizer was called with correct arguments
    mock_reranker._tokenizer.assert_called_once()
    call_args = mock_reranker._tokenizer.call_args
    
    # First argument should be list of queries
    assert call_args[0][0] == [query, query]
    # Second argument should be list of documents
    assert call_args[0][1] == documents
    # Check keyword arguments
    assert call_args[1]["return_tensors"] == "pt"
    assert call_args[1]["truncation"] is True
    assert call_args[1]["padding"] is True


@pytest.mark.unit
def test_call_model_called_with_tokenizer_output(
    mock_reranker: Reranker, mocker: MockerFixture
):
    """
    Test that the model is called with the tokenizer"s output.
    Args:
        mock_reranker (Reranker): The Reranker instance with mocked dependencies.
        mocker (MockerFixture): The pytest-mock fixture for mocking.
    """
    
    query = "test"
    documents = ["document"]
    
    # Set up tokenizer mock output
    tokenizer_output = {
        "input_ids": torch.tensor([[1, 2, 3]]),
        "attention_mask": torch.tensor([[1, 1, 1]])
    }
    mock_reranker._tokenizer.return_value = tokenizer_output
    
    # Mock model output
    mock_logits = torch.tensor([[0.5]])
    mock_output = mocker.MagicMock()
    mock_output.logits = mock_logits
    mock_reranker._model.return_value = mock_output
    
    mock_reranker(query, documents)
    
    # Verify model was called with tokenizer output
    mock_reranker._model.assert_called_once()
    call_kwargs = mock_reranker._model.call_args[1]
    assert "input_ids" in call_kwargs or len(mock_reranker._model.call_args[0]) > 0


@pytest.mark.unit
def test_call_with_empty_document_list(
    mock_reranker: Reranker, mocker: MockerFixture
):
    """
    Test that __call__ handles an empty document list correctly.
    Args:
        mock_reranker (Reranker): The Reranker instance with mocked dependencies.
        mocker (MockerFixture): The pytest-mock fixture for mocking.
    """
    
    query = "test query"
    documents = []
    
    # Mock model output for empty list
    mock_logits = torch.tensor([]).reshape(0, 1)
    mock_output = mocker.MagicMock()
    mock_output.logits = mock_logits
    mock_reranker._model.return_value = mock_output
    
    result = mock_reranker(query, documents)
    
    assert isinstance(result, list)
    assert len(result) == 0


@pytest.mark.unit
def test_call_converts_single_string_to_list(
    mock_reranker: Reranker, mocker: MockerFixture
):
    """
    Test that a single document string is converted to a list internally.
    Args:
    mock_reranker (Reranker): The Reranker instance with mocked dependencies.
        mocker (MockerFixture): The pytest-mock fixture for mocking.
    """
    
    query = "query"
    document = "single document"
    
    # Mock model output
    mock_logits = torch.tensor([[0.8]])
    mock_output = mocker.MagicMock()
    mock_output.logits = mock_logits
    mock_reranker._model.return_value = mock_output
    
    result = mock_reranker(query, document)
    
    # Result should be a list with one element
    assert isinstance(result, list)
    assert len(result) == 1
    assert result[0][0] == document


@pytest.mark.unit
def test_call_returns_tuples_with_correct_types(
    mock_reranker: Reranker, mocker: MockerFixture
):
    """
    Test that the return value contains tuples of (str, float).
    Args:
        mock_reranker (Reranker): The Reranker instance with mocked dependencies.
        mocker (MockerFixture): The pytest-mock fixture for mocking.
    """
    query = "test"
    documents = ["doc1", "doc2", "doc3"]
    
    # Mock model output
    mock_logits = torch.tensor([[0.5], [0.7], [0.3]])
    mock_output = mocker.MagicMock()
    mock_output.logits = mock_logits
    mock_reranker._model.return_value = mock_output
    
    result = mock_reranker(query, documents)
    
    for item in result:
        assert isinstance(item, tuple)
        assert len(item) == 2
        assert isinstance(item[0], str)
        assert isinstance(item[1], float)


@pytest.mark.unit
def test_call_logs_inference_with_decorator(
    mock_reranker: Reranker,
    mocker: MockerFixture,
    tmp_path
):
    """
    Test that __call__ logs inference metrics through the @track_inference_calls decorator.
    
    Args:
        mock_reranker (Reranker): The Reranker instance with mocked dependencies.
        mocker (MockerFixture): The pytest-mock fixture for mocking.
        tmp_path: Pytest fixture for temporary directory.
    """
    import json
    from pathlib import Path
    
    log_file = tmp_path / "inference.log"
    
    # Mock the config paths and decorator dependencies
    mocker.patch("artifex.core.decorators.logging.config.INFERENCE_LOGS_PATH", str(log_file))
    mocker.patch("artifex.core.decorators.logging._calculate_daily_inference_aggregates")
    
    # Mock psutil to avoid system calls
    mocker.patch("artifex.core.decorators.logging.psutil.virtual_memory", return_value=mocker.MagicMock(percent=50.0))
    mock_process = mocker.MagicMock()
    mock_process.cpu_percent.return_value = 25.0
    mocker.patch("artifex.core.decorators.logging.psutil.Process", return_value=mock_process)
    mocker.patch("artifex.core.decorators.logging.psutil.cpu_count", return_value=4)
    mocker.patch("artifex.core.decorators.logging.time.time", side_effect=[100.0, 101.0])
    
    query = "test query"
    documents = ["doc1", "doc2", "doc3"]
    
    # Mock model output
    mock_logits = torch.tensor([[0.5], [0.7], [0.3]])
    mock_output = mocker.MagicMock()
    mock_output.logits = mock_logits
    mock_reranker._model.return_value = mock_output
    
    # Call the method
    result = mock_reranker(query, documents)
    
    # Verify the log file was created
    assert log_file.exists()
    
    # Read and verify log entry
    log_content = log_file.read_text().strip()
    log_entry = json.loads(log_content)
    
    # Verify log entry contains expected fields
    assert log_entry["entry_type"] == "inference"
    assert log_entry["model"] == "Reranker"
    assert "inputs" in log_entry
    assert "output" in log_entry
    assert "inference_duration_seconds" in log_entry
    assert "cpu_usage_percent" in log_entry
    assert "ram_usage_percent" in log_entry
    assert "input_token_count" in log_entry
    assert "timestamp" in log_entry
    
    # Verify result is correct (should be sorted by score descending)
    assert len(result) == 3
    assert all(isinstance(item, tuple) for item in result)
    assert all(isinstance(item[0], str) and isinstance(item[1], float) for item in result)


@pytest.mark.unit
def test_call_with_disable_logging_prevents_logging(
    mock_reranker: Reranker,
    mocker: MockerFixture,
    tmp_path
):
    """
    Test that __call__ does not log when disable_logging=True is passed.
    
    Args:
        mock_reranker (Reranker): The Reranker instance with mocked dependencies.
        mocker (MockerFixture): The pytest-mock fixture for mocking.
        tmp_path: Pytest fixture for temporary directory.
    """
    import json
    from pathlib import Path
    
    log_file = tmp_path / "inference.log"
    
    # Mock the config paths
    mocker.patch("artifex.core.decorators.logging.config.INFERENCE_LOGS_PATH", str(log_file))
    
    query = "search query"
    documents = ["document1", "document2"]
    
    # Mock model output
    mock_logits = torch.tensor([[0.6], [0.4]])
    mock_output = mocker.MagicMock()
    mock_output.logits = mock_logits
    mock_reranker._model.return_value = mock_output
    
    # Call the method with disable_logging=True
    result = mock_reranker(query, documents, disable_logging=True)
    
    # Verify the log file was NOT created
    assert not log_file.exists()
    
    # Verify result is still correct
    assert len(result) == 2
    assert all(isinstance(item, tuple) for item in result)
    assert all(isinstance(item[0], str) and isinstance(item[1], float) for item in result)


import pytest
from pytest_mock import MockerFixture
from artifex.models import ClassificationModel


def make_mock_model(mocker: MockerFixture, id2label=None) -> MockerFixture:
    mock_model = mocker.Mock()
    mock_config = mocker.Mock()
    mock_config.id2label = id2label
    mock_model.config = mock_config
    return mock_model


@pytest.fixture
def mock_classification_model(mocker: MockerFixture) -> ClassificationModel:
    """
    Fixture to create a ClassificationModel with all dependencies mocked.
    """
    
    mock_synthex = mocker.Mock()
    # Patch Hugging Face model/tokenizer loading for __init__
    mock_model = make_mock_model(mocker, {0: "labelA", 1: "labelB"})
    mocker.patch(
        "artifex.models.classification.classification_model.AutoModelForSequenceClassification.from_pretrained",
        return_value=mock_model
    )
    mocker.patch(
        "artifex.models.classification.classification_model.AutoTokenizer.from_pretrained",
        return_value=mocker.Mock()
    )
    mocker.patch(
        "artifex.models.classification.classification_model.ClassLabel",
        return_value=mocker.Mock(names=["labelA", "labelB"])
    )
    mocker.patch("artifex.models.base_model.BaseModel.__init__", return_value=None)
    return ClassificationModel(mock_synthex)


def test_load_model_sets_model_and_labels(
    mocker: MockerFixture, mock_classification_model: ClassificationModel
):
    """
    Test that _load_model sets the model and labels correctly.
    Args:
        mocker (MockerFixture): The pytest-mock fixture for mocking.
        mock_classification_model (ClassificationModel): The mocked ClassificationModel instance.
    """
    
    # Prepare a mock model with id2label
    id2label = {0: "foo", 1: "bar"}
    mock_model = make_mock_model(mocker, id2label)
    mock_classlabel = mocker.Mock(names=["foo", "bar"])
    mocker.patch(
        "artifex.models.classification.classification_model.AutoModelForSequenceClassification.from_pretrained",
        return_value=mock_model
    )
    classlabel_patch = mocker.patch(
        "artifex.models.classification.classification_model.ClassLabel",
        return_value=mock_classlabel
    )

    mock_classification_model._load_model("dummy_path")

    # Model should be set
    assert mock_classification_model._model is mock_model
    # ClassLabel should be called with correct names
    classlabel_patch.assert_called_once_with(names=["foo", "bar"])
    # _labels should be set to the mock_classlabel
    assert mock_classification_model._labels is mock_classlabel


def test_load_model_raises_if_id2label_missing(
    mocker: MockerFixture, mock_classification_model: ClassificationModel
):
    """
    Test that _load_model raises AssertionError if id2label is missing.
    Args:
        mocker (MockerFixture): The pytest-mock fixture for mocking.
        mock_classification_model (ClassificationModel): The mocked ClassificationModel instance.
    """
    
    # Prepare a mock model with no id2label
    mock_model = make_mock_model(mocker, id2label=None)
    mocker.patch(
        "artifex.models.classification.classification_model.AutoModelForSequenceClassification.from_pretrained",
        return_value=mock_model
    )
    with pytest.raises(AssertionError, match="Model config must have id2label mapping."):
        mock_classification_model._load_model("dummy_path")


def test_load_model_passes_path(
    mocker: MockerFixture, mock_classification_model: ClassificationModel
):
    """
    Test that _load_model passes the correct path to from_pretrained.
    Args:
        mocker (MockerFixture): The pytest-mock fixture for mocking.
        mock_classification_model (ClassificationModel): The mocked ClassificationModel instance.
    """
    
    id2label = {0: "a", 1: "b"}
    mock_model = make_mock_model(mocker, id2label)
    from_pretrained_patch = mocker.patch(
        "artifex.models.classification.classification_model.AutoModelForSequenceClassification.from_pretrained",
        return_value=mock_model
    )
    mocker.patch(
        "artifex.models.classification.classification_model.ClassLabel",
        return_value=mocker.Mock(names=["a", "b"])
    )
    path = "some/model/path"
    mock_classification_model._load_model(path)
    from_pretrained_patch.assert_called_once_with(path)


import pytest
import shutil

from artifex import Artifex


@pytest.mark.integration
def test_train_success(
    artifex: Artifex,
    output_folder: str
):
    """
    Test the `train` method of the `EmotionDetection` class. Ensure that:
    - The training process completes without errors.
    - The output model's id2label mapping is the expected one.
    - The output model's label2id mapping is the expected one.
    Args:
        artifex (Artifex): The Artifex instance to be used for testing.
    """
    
    ed = artifex.emotion_detection
        
    try:
        ed.train(
            domain="test domain",
            classes={
                "happy": "text expressing happiness",
                "sad": "text expressing sadness"
            },
            num_samples=40,
            num_epochs=1,
            output_path=output_folder,
            device=-1,
            language="english",
            disable_logging=True
        )
        
        # Verify the model's config mappings
        id2label = ed._model.config.id2label
        label2id = ed._model.config.label2id
        assert id2label == { 0: "happy", 1: "sad" }
        assert label2id == { "happy": 0, "sad": 1 }
    finally:
        # Clean up the output folder
        shutil.rmtree(output_folder, ignore_errors=True)
    
    


import pytest

from artifex import Artifex


@pytest.mark.integration
def test__call__single_input_success(
    artifex: Artifex
):
    """
    Test the `__call__` method of the `Reranker` class when a single input is provided. 
    Ensure that: 
    - The return type is list[tuple[str, float]].
    - The output tuples only contain the provided input documents
    Args:
        artifex (Artifex): The Artifex instance to be used for testing.
    """
    
    input_doc = "doc1"
    
    out = artifex.reranker(
        query="test query", documents=input_doc, disable_logging=True
    )
    assert isinstance(out, list)
    assert all(
        isinstance(resp, tuple) and 
        isinstance(resp[0], str) and 
        isinstance(resp[1], float) 
        for resp in out
    )
    assert all(resp[0] in [input_doc] for resp in out)
    
@pytest.mark.integration
def test__call__multiple_inputs_success(
    artifex: Artifex
):
    """
    Test the `__call__` method of the `Reranker` class when multiple inputs are provided. 
    Ensure that: 
    - The return type is list[tuple[str, float]].
    - The output tuples only contain the provided input documents.
    Args:
        artifex (Artifex): The Artifex instance to be used for testing.
    """
    
    input_docs = ["doc1", "doc2", "doc3"]
    
    out = artifex.reranker(
        query="test query", documents=input_docs, disable_logging=True
    )
    assert isinstance(out, list)
    assert all(
        isinstance(resp, tuple) and 
        isinstance(resp[0], str) and 
        isinstance(resp[1], float) 
        for resp in out
    )
    assert all(resp[0] in input_docs for resp in out)


import pytest
from pytest_mock import MockerFixture
from synthex import Synthex
from unittest.mock import MagicMock

from artifex.models.named_entity_recognition import NamedEntityRecognition
from artifex.core import ParsedModelInstructions
from artifex.config import config


@pytest.fixture
def mock_dependencies(mocker: MockerFixture) -> None:
    """
    Fixture to mock external dependencies for NamedEntityRecognition.
    
    Args:
        mocker (MockerFixture): The pytest-mock fixture for mocking.
    """
    
    mocker.patch(
        'artifex.models.named_entity_recognition.named_entity_recognition.AutoModelForTokenClassification.from_pretrained',
        return_value=MagicMock()
    )
    mocker.patch(
        'artifex.models.named_entity_recognition.named_entity_recognition.AutoTokenizer.from_pretrained',
        return_value=MagicMock()
    )


@pytest.fixture
def mock_synthex(mocker: MockerFixture) -> Synthex:
    """
    Fixture to create a mock Synthex instance.
    
    Args:
        mocker (MockerFixture): The pytest-mock fixture for mocking.
    
    Returns:
        Synthex: A mocked Synthex instance.
    """
    
    return mocker.MagicMock()


@pytest.fixture
def ner(mock_dependencies: None, mock_synthex: Synthex) -> NamedEntityRecognition:
    """
    Fixture to create a NamedEntityRecognition instance for testing.
    
    Args:
        mock_dependencies (None): Fixture that mocks external dependencies.
        mock_synthex (Synthex): A mocked Synthex instance.
    
    Returns:
        NamedEntityRecognition: A NamedEntityRecognition instance.
    """
    
    return NamedEntityRecognition(synthex=mock_synthex)


@pytest.mark.unit
def test_get_data_gen_instr_returns_list_of_strings(
    ner: NamedEntityRecognition
) -> None:
    """
    Test that _get_data_gen_instr returns a list of strings.
    
    Args:
        ner (NamedEntityRecognition): The NamedEntityRecognition instance.
    """
    
    user_instr = ParsedModelInstructions(
        user_instructions=["PERSON: People names"],
        domain="News articles",
        language="english"
    )
    
    result = ner._get_data_gen_instr(user_instr)
    
    assert isinstance(result, list)
    assert all(isinstance(item, str) for item in result)


@pytest.mark.unit
def test_get_data_gen_instr_formats_domain_placeholder(
    ner: NamedEntityRecognition
) -> None:
    """
    Test that _get_data_gen_instr correctly formats the domain placeholder.
    
    Args:
        ner (NamedEntityRecognition): The NamedEntityRecognition instance.
    """
    
    user_instr = ParsedModelInstructions(
        user_instructions=["PERSON: People names"],
        domain="Medical records",
        language="english"
    )
    
    result = ner._get_data_gen_instr(user_instr)
    
    # Find instruction that contains domain
    domain_instructions = [instr for instr in result if "Medical records" in instr]
    assert len(domain_instructions) > 0


@pytest.mark.unit
def test_get_data_gen_instr_formats_language_placeholder(
    ner: NamedEntityRecognition
) -> None:
    """
    Test that _get_data_gen_instr correctly formats the language placeholder.
    
    Args:
        ner (NamedEntityRecognition): The NamedEntityRecognition instance.
    """
    
    user_instr = ParsedModelInstructions(
        user_instructions=["PERSON: People names"],
        domain="News",
        language="spanish"
    )
    
    result = ner._get_data_gen_instr(user_instr)
    
    # Find instruction that contains language
    language_instructions = [instr for instr in result if "spanish" in instr]
    assert len(language_instructions) > 0


@pytest.mark.unit
def test_get_data_gen_instr_formats_named_entity_tags_placeholder(
    ner: NamedEntityRecognition
) -> None:
    """
    Test that _get_data_gen_instr correctly formats the named_entity_tags placeholder.
    
    Args:
        ner (NamedEntityRecognition): The NamedEntityRecognition instance.
    """
    
    user_instr = ParsedModelInstructions(
        user_instructions=["PERSON: People names", "LOCATION: Geographic locations"],
        domain="News",
        language="english"
    )
    
    result = ner._get_data_gen_instr(user_instr)
    
    # Find instruction that contains named entity tags
    result_str = " ".join(result)
    assert "PERSON: People names" in result_str or "LOCATION: Geographic locations" in result_str


@pytest.mark.unit
def test_get_data_gen_instr_returns_correct_number_of_instructions(
    ner: NamedEntityRecognition
) -> None:
    """
    Test that _get_data_gen_instr returns same number of instructions as system instructions.
    
    Args:
        ner (NamedEntityRecognition): The NamedEntityRecognition instance.
    """
    
    user_instr = ParsedModelInstructions(
        user_instructions=["PERSON: People names"],
        domain="News",
        language="english"
    )
    
    result = ner._get_data_gen_instr(user_instr)
    
    assert len(result) == len(ner._system_data_gen_instr)


@pytest.mark.unit
def test_get_data_gen_instr_with_single_entity_tag(
    ner: NamedEntityRecognition
) -> None:
    """
    Test _get_data_gen_instr with a single named entity tag.
    
    Args:
        ner (NamedEntityRecognition): The NamedEntityRecognition instance.
    """
    
    user_instr = ParsedModelInstructions(
        user_instructions=["ORGANIZATION: Company names"],
        domain="Business",
        language="english"
    )
    
    result = ner._get_data_gen_instr(user_instr)
    
    result_str = " ".join(result)
    assert "ORGANIZATION: Company names" in result_str


@pytest.mark.unit
def test_get_data_gen_instr_with_multiple_entity_tags(
    ner: NamedEntityRecognition
) -> None:
    """
    Test _get_data_gen_instr with multiple named entity tags.
    
    Args:
        ner (NamedEntityRecognition): The NamedEntityRecognition instance.
    """
    
    user_instr = ParsedModelInstructions(
        user_instructions=[
            "PERSON: People names",
            "LOCATION: Geographic locations",
            "ORGANIZATION: Companies"
        ],
        domain="News",
        language="english"
    )
    
    result = ner._get_data_gen_instr(user_instr)
    
    result_str = " ".join(result)
    assert "PERSON: People names" in result_str
    assert "LOCATION: Geographic locations" in result_str
    assert "ORGANIZATION: Companies" in result_str


@pytest.mark.unit
def test_get_data_gen_instr_with_different_languages(
    ner: NamedEntityRecognition
) -> None:
    """
    Test _get_data_gen_instr with different language values.
    
    Args:
        ner (NamedEntityRecognition): The NamedEntityRecognition instance.
    """
    
    for language in ["english", "spanish", "french", "german", "chinese"]:
        user_instr = ParsedModelInstructions(
            user_instructions=["PERSON: People names"],
            domain="News",
            language=language
        )
        
        result = ner._get_data_gen_instr(user_instr)
        
        assert any(language in instr for instr in result)


@pytest.mark.unit
def test_get_data_gen_instr_with_unicode_in_domain(
    ner: NamedEntityRecognition
) -> None:
    """
    Test _get_data_gen_instr with unicode characters in domain.
    
    Args:
        ner (NamedEntityRecognition): The NamedEntityRecognition instance.
    """
    
    user_instr = ParsedModelInstructions(
        user_instructions=["PERSON: Personas"],
        domain="Artículos de noticias 新闻",
        language="spanish"
    )
    
    result = ner._get_data_gen_instr(user_instr)
    
    result_str = " ".join(result)
    assert "Artículos de noticias 新闻" in result_str or "spanish" in result_str


@pytest.mark.unit
def test_get_data_gen_instr_with_unicode_language(
    ner: NamedEntityRecognition
) -> None:
    """
    Test _get_data_gen_instr with unicode language parameter.
    
    Args:
        ner (NamedEntityRecognition): The NamedEntityRecognition instance.
    """
    
    user_instr = ParsedModelInstructions(
        user_instructions=["PERSON: 人名"],
        domain="News",
        language="中文"
    )
    
    result = ner._get_data_gen_instr(user_instr)
    
    assert any("中文" in instr for instr in result)


@pytest.mark.unit
def test_get_data_gen_instr_with_long_domain_string(
    ner: NamedEntityRecognition
) -> None:
    """
    Test _get_data_gen_instr with long domain string.
    
    Args:
        ner (NamedEntityRecognition): The NamedEntityRecognition instance.
    """
    
    long_domain = "Medical and healthcare records including patient information, doctor notes, prescriptions, and clinical trial data"
    
    user_instr = ParsedModelInstructions(
        user_instructions=["PERSON: Patient names"],
        domain=long_domain,
        language="english"
    )
    
    result = ner._get_data_gen_instr(user_instr)
    
    assert any(long_domain in instr for instr in result)


@pytest.mark.unit
def test_get_data_gen_instr_with_special_characters_in_entity_tags(
    ner: NamedEntityRecognition
) -> None:
    """
    Test _get_data_gen_instr with special characters in entity tag descriptions.
    
    Args:
        ner (NamedEntityRecognition): The NamedEntityRecognition instance.
    """
    
    user_instr = ParsedModelInstructions(
        user_instructions=["PERSON: Names (first & last)", "ORG: Companies, Inc."],
        domain="Business",
        language="english"
    )
    
    result = ner._get_data_gen_instr(user_instr)
    
    result_str = " ".join(result)
    assert "Names (first & last)" in result_str or "Companies, Inc." in result_str


@pytest.mark.unit
def test_get_data_gen_instr_does_not_modify_input(
    ner: NamedEntityRecognition
) -> None:
    """
    Test that _get_data_gen_instr does not modify the input ParsedModelInstructions.
    
    Args:
        ner (NamedEntityRecognition): The NamedEntityRecognition instance.
    """
    
    user_instr = ParsedModelInstructions(
        user_instructions=["PERSON: People", "LOCATION: Places"],
        domain="News",
        language="english"
    )
    
    original_instructions = user_instr.user_instructions.copy()
    original_domain = user_instr.domain
    original_language = user_instr.language
    
    ner._get_data_gen_instr(user_instr)
    
    assert user_instr.user_instructions == original_instructions
    assert user_instr.domain == original_domain
    assert user_instr.language == original_language


@pytest.mark.unit
def test_get_data_gen_instr_with_empty_entity_tags_list(
    ner: NamedEntityRecognition
) -> None:
    """
    Test _get_data_gen_instr with empty user_instructions list.
    
    Args:
        ner (NamedEntityRecognition): The NamedEntityRecognition instance.
    """
    
    user_instr = ParsedModelInstructions(
        user_instructions=[],
        domain="News",
        language="english"
    )
    
    result = ner._get_data_gen_instr(user_instr)
    
    # Should still return formatted system instructions
    assert len(result) == len(ner._system_data_gen_instr)
    assert all(isinstance(item, str) for item in result)


@pytest.mark.unit
def test_get_data_gen_instr_with_whitespace_in_domain(
    ner: NamedEntityRecognition
) -> None:
    """
    Test _get_data_gen_instr with whitespace in domain.
    
    Args:
        ner (NamedEntityRecognition): The NamedEntityRecognition instance.
    """
    
    user_instr = ParsedModelInstructions(
        user_instructions=["PERSON: People"],
        domain="  News articles  ",
        language="english"
    )
    
    result = ner._get_data_gen_instr(user_instr)
    
    result_str = " ".join(result)
    assert "  News articles  " in result_str or "News articles" in result_str


@pytest.mark.unit
def test_get_data_gen_instr_with_quotes_in_entity_descriptions(
    ner: NamedEntityRecognition
) -> None:
    """
    Test _get_data_gen_instr with quotes in entity descriptions.
    
    Args:
        ner (NamedEntityRecognition): The NamedEntityRecognition instance.
    """
    
    user_instr = ParsedModelInstructions(
        user_instructions=['PERSON: Names like "John Doe"', "TITLE: Books with 'quotes'"],
        domain="Literature",
        language="english"
    )
    
    result = ner._get_data_gen_instr(user_instr)
    
    result_str = " ".join(result)
    assert '"John Doe"' in result_str or "'quotes'" in result_str


@pytest.mark.unit
def test_get_data_gen_instr_with_newlines_in_entity_descriptions(
    ner: NamedEntityRecognition
) -> None:
    """
    Test _get_data_gen_instr with newlines in entity descriptions.
    
    Args:
        ner (NamedEntityRecognition): The NamedEntityRecognition instance.
    """
    
    user_instr = ParsedModelInstructions(
        user_instructions=["PERSON: People\nwith\nnewlines"],
        domain="News",
        language="english"
    )
    
    result = ner._get_data_gen_instr(user_instr)
    
    assert any("PERSON: People\\nwith\\nnewlines" in instr for instr in result)


@pytest.mark.unit
def test_get_data_gen_instr_with_numeric_strings(
    ner: NamedEntityRecognition
) -> None:
    """
    Test _get_data_gen_instr with numeric strings in descriptions.
    
    Args:
        ner (NamedEntityRecognition): The NamedEntityRecognition instance.
    """
    
    user_instr = ParsedModelInstructions(
        user_instructions=["CODE: Numbers like 123", "ID: 456.789"],
        domain="Technical",
        language="english"
    )
    
    result = ner._get_data_gen_instr(user_instr)
    
    result_str = " ".join(result)
    assert "123" in result_str and "456.789" in result_str


@pytest.mark.unit
def test_get_data_gen_instr_formats_all_system_instructions(
    ner: NamedEntityRecognition
) -> None:
    """
    Test that _get_data_gen_instr formats all system instructions.
    
    Args:
        ner (NamedEntityRecognition): The NamedEntityRecognition instance.
    """
    
    user_instr = ParsedModelInstructions(
        user_instructions=["PERSON: People"],
        domain="News",
        language="english"
    )
    
    result = ner._get_data_gen_instr(user_instr)
    
    # Result should have same length as system instructions
    assert len(result) == len(ner._system_data_gen_instr)
    
    # All should be non-empty strings
    assert all(len(instr) > 0 for instr in result)


@pytest.mark.unit
def test_get_data_gen_instr_with_mixed_case_language(
    ner: NamedEntityRecognition
) -> None:
    """
    Test _get_data_gen_instr with mixed case language parameter.
    
    Args:
        ner (NamedEntityRecognition): The NamedEntityRecognition instance.
    """
    
    user_instr = ParsedModelInstructions(
        user_instructions=["PERSON: People"],
        domain="News",
        language="EnGLisH"
    )
    
    result = ner._get_data_gen_instr(user_instr)
    
    assert any("EnGLisH" in instr for instr in result)


@pytest.mark.unit
def test_get_data_gen_instr_with_complex_punctuation(
    ner: NamedEntityRecognition
) -> None:
    """
    Test _get_data_gen_instr with complex punctuation in descriptions.
    
    Args:
        ner (NamedEntityRecognition): The NamedEntityRecognition instance.
    """
    
    user_instr = ParsedModelInstructions(
        user_instructions=["PERSON: Names (with [nested {punctuation}])"],
        domain="News",
        language="english"
    )
    
    result = ner._get_data_gen_instr(user_instr)
    
    assert any("Names (with [nested {punctuation}])" in instr for instr in result)


@pytest.mark.unit
def test_get_data_gen_instr_with_emoji_in_descriptions(
    ner: NamedEntityRecognition
) -> None:
    """
    Test _get_data_gen_instr with emoji characters in descriptions.
    
    Args:
        ner (NamedEntityRecognition): The NamedEntityRecognition instance.
    """
    
    user_instr = ParsedModelInstructions(
        user_instructions=["PERSON: People 👤", "LOCATION: Places 🌍"],
        domain="Social media",
        language="english"
    )
    
    result = ner._get_data_gen_instr(user_instr)
    
    result_str = " ".join(result)
    assert "👤" in result_str or "🌍" in result_str


@pytest.mark.unit
def test_get_data_gen_instr_result_contains_only_strings(
    ner: NamedEntityRecognition
) -> None:
    """
    Test that _get_data_gen_instr result contains only string types.
    
    Args:
        ner (NamedEntityRecognition): The NamedEntityRecognition instance.
    """
    
    user_instr = ParsedModelInstructions(
        user_instructions=["PERSON: People", "LOCATION: Places"],
        domain="News",
        language="english"
    )
    
    result = ner._get_data_gen_instr(user_instr)
    
    for item in result:
        assert isinstance(item, str)
        assert not isinstance(item, (list, dict, tuple))


@pytest.mark.unit
def test_get_data_gen_instr_with_backslashes(
    ner: NamedEntityRecognition
) -> None:
    """
    Test _get_data_gen_instr with backslashes in descriptions.
    
    Args:
        ner (NamedEntityRecognition): The NamedEntityRecognition instance.
    """
    
    user_instr = ParsedModelInstructions(
        user_instructions=["PATH: Paths\\with\\backslashes"],
        domain="Technical",
        language="english"
    )
    
    result = ner._get_data_gen_instr(user_instr)
    
    assert any("Paths\\\\with\\\\backslashes" in instr for instr in result)


@pytest.mark.unit
def test_get_data_gen_instr_consecutive_calls_produce_same_result(
    ner: NamedEntityRecognition
) -> None:
    """
    Test that consecutive calls to _get_data_gen_instr with same input produce same result.
    
    Args:
        ner (NamedEntityRecognition): The NamedEntityRecognition instance.
    """
    
    user_instr = ParsedModelInstructions(
        user_instructions=["PERSON: People", "LOCATION: Places"],
        domain="News",
        language="english"
    )
    
    result1 = ner._get_data_gen_instr(user_instr)
    result2 = ner._get_data_gen_instr(user_instr)
    
    assert result1 == result2


@pytest.mark.unit
def test_get_data_gen_instr_with_colon_in_descriptions(
    ner: NamedEntityRecognition
) -> None:
    """
    Test _get_data_gen_instr with colons in entity descriptions.
    
    Args:
        ner (NamedEntityRecognition): The NamedEntityRecognition instance.
    """
    
    user_instr = ParsedModelInstructions(
        user_instructions=["PERSON: Names like: John, Jane, etc."],
        domain="News",
        language="english"
    )
    
    result = ner._get_data_gen_instr(user_instr)
    
    assert any("Names like: John, Jane, etc." in instr for instr in result)


@pytest.mark.unit
def test_get_data_gen_instr_with_tabs_in_descriptions(
    ner: NamedEntityRecognition
) -> None:
    """
    Test _get_data_gen_instr with tab characters in descriptions.
    
    Args:
        ner (NamedEntityRecognition): The NamedEntityRecognition instance.
    """
    
    user_instr = ParsedModelInstructions(
        user_instructions=["PERSON: Names\twith\ttabs"],
        domain="News",
        language="english"
    )
    
    result = ner._get_data_gen_instr(user_instr)
    
    assert any("Names\\twith\\ttabs" in instr for instr in result)


@pytest.mark.unit
def test_get_data_gen_instr_uses_user_instructions_field_for_tags(
    ner: NamedEntityRecognition
) -> None:
    """
    Test that _get_data_gen_instr uses user_instructions field for named entity tags.
    
    Args:
        ner (NamedEntityRecognition): The NamedEntityRecognition instance.
    """
    
    user_instr = ParsedModelInstructions(
        user_instructions=["PERSON: People names", "ORGANIZATION: Company names"],
        domain="Business news",
        language="english"
    )
    
    result = ner._get_data_gen_instr(user_instr)
    
    result_str = " ".join(result)
    # Should use user_instructions for named entity tags
    assert "PERSON: People names" in result_str or "ORGANIZATION: Company names" in result_str
